#ifndef NB_DEVICE_HPP
#define NB_DEVICE_HPP

#include <NBDevicesTypes.hpp>
#include <iterator>
#include <vector>
#include <cstring>

namespace NextBiometrics
{
    /* NextBiometrics.Devices namespace */
    namespace Devices
    {
#include <NBDevice.h>

#undef NB_DEVICE_SCAN_TIMEOUT_INFINITE
#undef NB_DEVICE_SCAN_SKIP_FINGER_NOT_REMOVED_STATUS_FLAG
#undef NB_DEVICE_CONNECT_TO_SPI_SKIP_GPIO_INIT_FLAG

        class NBDeviceScanResult
        {
            friend class NBDevice;

            NBDeviceScanStatus m_eScanStatus;
            NBDeviceScanFormatInfo m_sScanFormatInfo;
            vector<NBByte> m_pImage;

            NBDeviceScanResult(NBDeviceScanStatus eScanStatus, const NBDeviceScanFormatInfo& sScanFormatInfo, const NBByte * pImage, NBSizeType stImageSize)
                : m_eScanStatus(eScanStatus), m_sScanFormatInfo(sScanFormatInfo)
            {
                if (pImage && stImageSize > 0)
                {
                    m_pImage.assign(pImage, pImage + stImageSize);
                }
            }
        public:
            NBDeviceScanResult() :
                NBDeviceScanResult(NBDeviceScanStatusNone, NBDeviceScanFormatInfo(), nullptr, 0)
            {
            }

            NBDeviceScanStatus GetStatus() const
            {
                return m_eScanStatus;
            }

            NBDeviceScanFormatInfo GetFormat() const
            {
                return m_sScanFormatInfo;
            }

            const NBByte * GetImage() const
            {
                return m_pImage.data();
            }

            NBSizeType GetImageSize() const
            {
                return m_pImage.size();
            }
        };

        class NBDeviceScanPreviewEventArgs;
        typedef void(*NBDeviceScanPreviewEvent)(NBDeviceScanPreviewEventArgs &args);

        class NBDevice
        {
            friend class NBDevices;
            friend class NBDevicesDeviceChangedEventArgs;
            friend class NBDeviceScanPreviewEventArgs;

            class NBDeviceScanPreviewEventHandler
            {
                friend class NBDevice;

                NBDeviceScanPreviewEvent m_callback;
                void * m_pParam;

                NBDeviceScanPreviewEventHandler(NBDeviceScanPreviewEvent callback, void * pParam)
                    : m_callback(callback), m_pParam(pParam)
                {
                }

                static NBResult NB_API NativeCallback(const HNBDevice hDevice, NBDeviceScanStatus * peStatus, const NBDeviceScanFormatInfo * psScanFormatInfo, const NBByte * pBuffer, NBSizeType stBufferSize, const NBDeviceScanPreviewDetails * psDetails, void * pParam);
            };

            class NBDeviceHandle
            {
                HNBDevice m_hHandle;
                bool m_bOwnsHandle;
            public:
                NBDeviceHandle(HNBDevice hHandle, bool bOwnsHandle)
                    : m_hHandle(hHandle), m_bOwnsHandle(bOwnsHandle)
                {
                }

                ~NBDeviceHandle()
                {
                    if (m_bOwnsHandle && m_hHandle) NBErrors::RequireNoErr(NBDeviceDestroy(m_hHandle));
                }

                HNBDevice GetHandle() const
                {
                    return m_hHandle;
                }

                bool OwnsHandle() const
                {
                    return m_bOwnsHandle;
                }
            };

            shared_ptr<NBDeviceHandle> m_hHandle;

            void CheckHandle() const
            {
                if (!m_hHandle || !m_hHandle->GetHandle()) NBThrowException(NBErrors::NB_ERROR_ARGUMENT_NULL, "m_hHandle is NULL");
            }

            vector<NBAChar> GetIdStrA() const
            {
                NBUInt uiValueLength;
                NBErrors::RequireNoErr(NBDeviceGetIdA(GetHandle(), nullptr, 0, &uiValueLength));
                vector<NBAChar> szString(uiValueLength + 1);
                NBErrors::RequireNoErr(NBDeviceGetIdA(GetHandle(), szString.data(), NBUInt(szString.size()), &uiValueLength));
                return szString;
            }

        public:
            static const NBInt SCAN_TIMEOUT_INFINITE = -1;
            static const NBUInt SCAN_SKIP_FINGER_NOT_REMOVED_STATUS_FLAG = 0x08;
            static const NBUInt CONNECT_TO_SPI_SKIP_GPIO_INIT_FLAG = 0x02;

            NBDevice(HNBDevice hHandle, bool bOwnsHandle)
                : NBDevice()
            {
                m_hHandle = make_shared<NBDeviceHandle>(hHandle, bOwnsHandle);
            }

            NBDevice(HNBDevice hHandle)
                : NBDevice(hHandle, true)
            {
            }

            NBDevice()
            {
                m_hHandle = make_shared<NBDeviceHandle>(nullptr, false);
            }

            static NBDevice ConnectToSpi(const NBAChar * szSpiName, NBInt iAwakePin, NBInt iResetPin)
            {
                return ConnectToSpi(szSpiName, iAwakePin, iResetPin, NBUInt(0));
            }

            static NBDevice ConnectToSpi(const NBAChar * szSpiName, NBInt iAwakePin, NBInt iResetPin, NBUInt uiFlags)
            {
                HNBDevice hDevice = nullptr;
                NBErrors::RequireNoErr(NBDeviceConnectToSpiA(szSpiName, iAwakePin, iResetPin, uiFlags, &hDevice));
                return NBDevice(hDevice);
            }

            static NBDevice ConnectToSpi(const NBString& szSpiName, NBInt iAwakePin, NBInt iResetPin)
            {
                return ConnectToSpi(FromNBString(szSpiName), iAwakePin, iResetPin);
            }

            static NBDevice ConnectToSpi(const NBString& szSpiName, NBInt iAwakePin, NBInt iResetPin, NBUInt uiFlags)
            {
                return ConnectToSpi(FromNBString(szSpiName), iAwakePin, iResetPin, uiFlags);
            }

            static NBDevice ConnectToSpi(const NBAChar * szSpiName, NBInt iAwakePin, NBInt iResetPin, NBInt iChipSelectPin)
            {
                return ConnectToSpi(szSpiName, iAwakePin, iResetPin, iChipSelectPin, NBUInt(0));
            }

            static NBDevice ConnectToSpi(const NBAChar * szSpiName, NBInt iAwakePin, NBInt iResetPin, NBInt iChipSelectPin, NBUInt uiFlags)
            {
                HNBDevice hDevice = nullptr;
                NBErrors::RequireNoErr(NBDeviceConnectToSpiExA(szSpiName, iAwakePin, iResetPin, iChipSelectPin, uiFlags, &hDevice));
                return NBDevice(hDevice);
            }

            static NBDevice ConnectToSpi(const NBString& szSpiName, NBInt iAwakePin, NBInt iResetPin, NBInt iChipSelectPin)
            {
                return ConnectToSpi(szSpiName, iAwakePin, iResetPin, iChipSelectPin, NBUInt(0));
            }

            static NBDevice ConnectToSpi(const NBString& szSpiName, NBInt iAwakePin, NBInt iResetPin, NBInt iChipSelectPin, NBUInt uiFlags)
            {
                return ConnectToSpi(FromNBString(szSpiName), iAwakePin, iResetPin, iChipSelectPin, uiFlags);
            }

            HNBDevice GetHandle() const
            {
                CheckHandle();
                return m_hHandle->GetHandle();
            }

            NBDeviceState GetState() const
            {
                NBDeviceState eState = NBDeviceStateNotConnected;

                NBErrors::RequireNoErr(NBDeviceGetState(GetHandle(), &eState));
                return eState;
            }

            NBDeviceType GetType() const
            {
                NBDeviceType eType;
                NBErrors::RequireNoErr(NBDeviceGetType(GetHandle(), &eType));
                return eType;
            }

            NBDeviceConnectionType GetConnectionType() const
            {
                NBDeviceConnectionType eType;
                NBErrors::RequireNoErr(NBDeviceGetConnectionType(GetHandle(), &eType));
                return eType;
            }

            NBVersion GetFirmwareVersion() const
            {
                NBVersion sVersion;
                NBErrors::RequireNoErr(NBDeviceGetFirmwareVersion(GetHandle(), &sVersion));
                return sVersion;
            }

            NBString GetId() const
            {
                NBUInt uiValueLength;
                NBErrors::RequireNoErr(NBDeviceGetIdA(GetHandle(), nullptr, 0, &uiValueLength));
                vector<NBAChar> szString(uiValueLength + 1);
                NBErrors::RequireNoErr(NBDeviceGetIdA(GetHandle(), szString.data(), NBUInt(szString.size()), &uiValueLength));
                return ToNBString(szString.data(), uiValueLength);
            }

            NBString GetManufacturer() const
            {
                NBUInt uiValueLength;
                NBErrors::RequireNoErr(NBDeviceGetManufacturerA(GetHandle(), nullptr, 0, &uiValueLength));
                vector<NBAChar> szString(uiValueLength + 1);
                NBErrors::RequireNoErr(NBDeviceGetManufacturerA(GetHandle(), szString.data(), NBUInt(szString.size()), &uiValueLength));
                return ToNBString(szString.data(), uiValueLength);
            }

            NBString GetModel() const
            {
                NBUInt uiValueLength;
                NBErrors::RequireNoErr(NBDeviceGetModelA(GetHandle(), nullptr, 0, &uiValueLength));
                vector<NBAChar> szString(uiValueLength + 1);
                NBErrors::RequireNoErr(NBDeviceGetModelA(GetHandle(), szString.data(), NBUInt(szString.size()), &uiValueLength));
                return ToNBString(szString.data(), uiValueLength);
            }

            NBString GetSerialNumber() const
            {
                NBUInt uiValueLength;
                NBErrors::RequireNoErr(NBDeviceGetSerialNumberA(GetHandle(), nullptr, 0, &uiValueLength));
                vector<NBAChar> szString(uiValueLength + 1);
                NBErrors::RequireNoErr(NBDeviceGetSerialNumberA(GetHandle(), szString.data(), NBUInt(szString.size()), &uiValueLength));
                return ToNBString(szString.data(), uiValueLength);
            }

            NBString GetProduct() const
            {
                NBUInt uiValueLength;
                NBErrors::RequireNoErr(NBDeviceGetProductA(GetHandle(), nullptr, 0, &uiValueLength));
                vector<NBAChar> szString(uiValueLength + 1);
                NBErrors::RequireNoErr(NBDeviceGetProductA(GetHandle(), szString.data(), NBUInt(szString.size()), &uiValueLength));
                return ToNBString(szString.data(), uiValueLength);
            }

            void Reset() const
            {
                NBErrors::RequireNoErr(NBDeviceReset(GetHandle()));
            }

            bool isScanFormatSupported(NBDeviceScanFormat eScanFormat) const
            {
                NBBool value;
                NBErrors::RequireNoErr(NBDeviceIsScanFormatSupported(GetHandle(), eScanFormat, &value));
                return value != NBFalse;
            }

            NBDeviceScanFormatInfo GetScanFormatInfo(NBDeviceScanFormat eScanFormat) const
            {
                NBDeviceScanFormatInfo sScanFormatInfo;
                NBErrors::RequireNoErr(NBDeviceGetScanFormatInfo(GetHandle(), eScanFormat, &sScanFormatInfo));
                return sScanFormatInfo;
            }

            vector<NBDeviceScanFormatInfo> GetSupportedScanFormats() const
            {
                NBUInt uiFormatsCount;
                NBErrors::RequireNoErr(NBDeviceGetSupportedScanFormats(GetHandle(), nullptr, 0, &uiFormatsCount));
                vector<NBDeviceScanFormatInfo> arFormats(uiFormatsCount);
                if (uiFormatsCount > 0)
                {
                    NBUInt uiNewFormatsCount;
                    vector<NBDeviceScanFormat> areScanFormats(uiFormatsCount);
                    NBErrors::RequireNoErr(NBDeviceGetSupportedScanFormats(GetHandle(), areScanFormats.data(), NBUInt(areScanFormats.size()), &uiNewFormatsCount));
                    for (NBUInt i = 0; i < min(uiFormatsCount, uiNewFormatsCount); i++)
                    {
                        arFormats[i] = GetScanFormatInfo(areScanFormats[i]);
                    }
                }
                return arFormats;
            }

            void SetParameter(NBUInt uiParameter, NBInt iValue) const
            {
                NBErrors::RequireNoErr(NBDeviceSetParameter(GetHandle(), uiParameter, iValue));
            }

            NBInt GetParameter(NBUInt uiParameter) const
            {
                NBInt value;
                NBErrors::RequireNoErr(NBDeviceGetParameter(GetHandle(), uiParameter, &value));
                return value;
            }

            void SetBlobParameter(NBUInt uiParameter, const NBByte *pBlob, NBSizeType stBlobSize) const
            {
                NBErrors::RequireNoErr(NBDeviceSetBlobParameter(GetHandle(), uiParameter, pBlob, stBlobSize));
            }

			vector<NBByte> GetBlobParameter(NBUInt uiParameter) const
			{
				NBByte* _piBlob;
				NBByte** _ppBlob = &_piBlob;
				NBSizeType* _pstBlobSize;
				NBSizeType _stBlobSize;
				_pstBlobSize = &_stBlobSize;

				NBErrors::RequireNoErr(NBDeviceGetBlobParameter(GetHandle(), uiParameter, _ppBlob, _pstBlobSize));

				vector<NBByte> _ppBlobVec(*_ppBlob, *_ppBlob + (*_pstBlobSize));

				NBDeviceFree(_piBlob);
				return _ppBlobVec;
			}

            NBInt GetFingerDetectValue() const
            {
                return GetFingerDetectValue(NBDeviceFingerDetectTypeQuick);
            }

            NBInt GetFingerDetectValue(NBDeviceFingerDetectType eDetectType) const
            {
                NBInt value;
                NBErrors::RequireNoErr(NBDeviceGetFingerDetectValue(GetHandle(), eDetectType, &value));
                return value;
            }

            NBDeviceScanResult Scan() const
            {
                return Scan(GetSupportedScanFormats()[0], 0);
            }

            NBDeviceScanResult Scan(NBDeviceScanFormat eScanFormat) const
            {
                return Scan(eScanFormat, 0);
            }

            NBDeviceScanResult Scan(NBDeviceScanFormat eScanFormat, NBUInt uiFlags) const
            {
                const NBDeviceScanFormatInfo sScanFormatInfo = GetScanFormatInfo(eScanFormat);
                return Scan(sScanFormatInfo, uiFlags);
            }

            NBDeviceScanResult Scan(const NBDeviceScanFormatInfo& sScanFormatInfo) const
            {
                return Scan(sScanFormatInfo, 0);
            }

            NBDeviceScanResult Scan(const NBDeviceScanFormatInfo& sScanFormatInfo, NBUInt uiFlags) const
            {
                vector<NBByte> pImage(sScanFormatInfo.uiHeight*sScanFormatInfo.uiWidth);
                NBDeviceScanStatus eStatus;
                NBErrors::RequireNoErr(NBDeviceScan(GetHandle(), sScanFormatInfo.eScanFormat, pImage.data(), pImage.size(), uiFlags, &eStatus));
                return NBDeviceScanResult(eStatus, sScanFormatInfo, pImage.data(), pImage.size());
            }

            NBDeviceScanResult ScanEx() const
            {
                return ScanEx(SCAN_TIMEOUT_INFINITE);
            }


            NBDeviceScanResult ScanEx(NBInt iTimeout) const
            {
                return ScanEx(iTimeout, nullptr, nullptr);
            }

            NBDeviceScanResult ScanEx(NBInt iTimeout, const NBDeviceScanPreviewEvent pPreviewCallback, void * pParam) const
            {
                return ScanEx(GetSupportedScanFormats()[0], iTimeout, pPreviewCallback, pParam);
            }

            NBDeviceScanResult ScanEx(NBDeviceScanFormat eScanFormat) const
            {
                return ScanEx(eScanFormat, SCAN_TIMEOUT_INFINITE);
            }

            NBDeviceScanResult ScanEx(const NBDeviceScanFormatInfo& sScanFormatInfo) const
            {
                return ScanEx(sScanFormatInfo, SCAN_TIMEOUT_INFINITE);
            }

            NBDeviceScanResult ScanEx(NBDeviceScanFormat eScanFormat, const NBDeviceScanPreviewEvent pPreviewCallback, void * pParam) const
            {
                return ScanEx(eScanFormat, SCAN_TIMEOUT_INFINITE, pPreviewCallback, pParam);
            }

            NBDeviceScanResult ScanEx(const NBDeviceScanFormatInfo& sScanFormatInfo, const NBDeviceScanPreviewEvent pPreviewCallback, void * pParam) const
            {
                return ScanEx(sScanFormatInfo, SCAN_TIMEOUT_INFINITE, pPreviewCallback, pParam);
            }

            NBDeviceScanResult ScanEx(NBDeviceScanFormat scanFormat, NBInt iTimeout) const
            {
                return ScanEx(scanFormat, iTimeout, nullptr, nullptr);
            }

            NBDeviceScanResult ScanEx(const NBDeviceScanFormatInfo& sScanFormatInfo, NBInt iTimeout) const
            {
                return ScanEx(sScanFormatInfo, iTimeout, nullptr, nullptr);
            }

            NBDeviceScanResult ScanEx(NBDeviceScanFormat eScanFormat, NBInt iTimeout, const NBDeviceScanPreviewEvent pPreviewCallback, void * pParam) const
            {
                return ScanEx(eScanFormat, iTimeout, pPreviewCallback, pParam, 0);
            }

            NBDeviceScanResult ScanEx(NBDeviceScanFormat eScanFormat, NBInt iTimeout, const NBDeviceScanPreviewEvent pPreviewCallback, void * pParam, NBUInt uiFlags) const
            {
                const NBDeviceScanFormatInfo sScanFormatInfo = GetScanFormatInfo(eScanFormat);
                return ScanEx(sScanFormatInfo, iTimeout, pPreviewCallback, pParam, uiFlags);
            }

            NBDeviceScanResult ScanEx(const NBDeviceScanFormatInfo& sScanFormatInfo, NBInt iTimeout, const NBDeviceScanPreviewEvent pPreviewCallback, void * pParam) const
            {
                return ScanEx(sScanFormatInfo, iTimeout, pPreviewCallback, pParam, 0);
            }

            NBDeviceScanResult ScanEx(const NBDeviceScanFormatInfo& sScanFormatInfo, NBInt iTimeout, const NBDeviceScanPreviewEvent pPreviewCallback, void * pParam, NBUInt uiFlags) const
            {
                vector<NBUInt8> pImage(sScanFormatInfo.uiHeight*sScanFormatInfo.uiWidth);
                unique_ptr<NBDeviceScanPreviewEventHandler> pScanPreviewEvent(new NBDeviceScanPreviewEventHandler(pPreviewCallback, pParam));
                NBDeviceScanStatus eStatus;
                NBErrors::RequireNoErr(NBDeviceScanEx(GetHandle(), sScanFormatInfo.eScanFormat, iTimeout, pPreviewCallback ? NBDeviceScanPreviewEventHandler::NativeCallback : nullptr, pScanPreviewEvent.get(), pImage.data(), pImage.size(), uiFlags, &eStatus));
                return NBDeviceScanResult(eStatus, sScanFormatInfo, pImage.data(), pImage.size());
            }

            void CancelScan() const
            {
                NBErrors::RequireNoErr(NBDeviceCancelScan(GetHandle()));
            }

            void EnterStopMode(NBDeviceStopMode eStopMode) const
            {
                EnterStopMode(eStopMode, 0);
            }

            void EnterStopMode(NBDeviceStopMode eStopMode, NBUInt uiFingerDetectionCount) const
            {
                NBErrors::RequireNoErr(NBDeviceEnterStopMode(GetHandle(), eStopMode, uiFingerDetectionCount));
            }

            void SetLedState(NBDeviceLedState eLedState) const
            {
                NBErrors::RequireNoErr(NBDeviceSetLedState(GetHandle(), eLedState));
            }

            bool operator==(const NBDevice& other) const
            {
                auto szId = GetIdStrA();
                auto szId2 = other.GetIdStrA();
                return strcmp(szId.data(), szId2.data()) == 0;
            }

            bool operator!=(const NBDevice &other) const
            {
                return !(*this == other);
            }
        };

        class NBDeviceScanPreviewEventArgs : public EventArgs
        {
            friend class NBDevice;

            NBDevice m_device;
            NBDeviceScanStatus m_eScanStatus;
            NBDeviceScanFormatInfo m_sScanFormatInfo;
            const NBByte * m_pImage;
            NBSizeType m_stImageSize;
            NBDeviceScanPreviewDetails m_sScanPreviewDetails;

            NBDeviceScanPreviewEventArgs(const NBDevice& device, NBDeviceScanStatus eScanStatus, const NBDeviceScanFormatInfo& sScanFormatInfo, const NBUInt8 * pImage, NBSizeType stImageSize, const NBDeviceScanPreviewDetails& sScanPreviewDetails, void * pParam)
                : EventArgs(pParam), m_device(device), m_eScanStatus(eScanStatus), m_sScanFormatInfo(sScanFormatInfo), m_pImage(pImage), m_stImageSize(stImageSize), m_sScanPreviewDetails(sScanPreviewDetails)
            {
            }
        public:
            const NBDevice& GetDevice() const
            {
                return m_device;
            }

            NBDeviceScanStatus GetStatus() const
            {
                return m_eScanStatus;
            }

            void SetStatus(NBDeviceScanStatus eStatus)
            {
                m_eScanStatus = eStatus;
            }

            const NBDeviceScanFormatInfo& GetFormat() const
            {
                return m_sScanFormatInfo;
            }

            const NBByte * GetImage() const
            {
                return m_pImage;
            }

            NBSizeType GetImageSize() const
            {
                return m_stImageSize;
            }

            NBInt GetFingerDetectValue() const
            {
                return m_sScanPreviewDetails.iFingerDetectValue;
            }
        };

        inline NBResult NB_API NBDevice::NBDeviceScanPreviewEventHandler::NativeCallback(const HNBDevice hDevice, NBDeviceScanStatus * peStatus, const NBDeviceScanFormatInfo * psScanFormatInfo, const NBByte * pBuffer, NBSizeType stBufferSize, const NBDeviceScanPreviewDetails * psDetails, void * pParam)
        {
            NBResult result = NBErrors::NB_OK;
            try
            {
                NBDeviceScanPreviewEventHandler * pHandler = reinterpret_cast<NBDeviceScanPreviewEventHandler*>(pParam);
                NBDeviceScanPreviewEventArgs args(NBDevice(hDevice, false), *peStatus, *psScanFormatInfo, pBuffer, stBufferSize, *psDetails, pHandler->m_pParam);
                pHandler->m_callback(args);
                *peStatus = args.GetStatus();
            }
            catch (const NBError &error)
            {
                result = NBErrors::SetLast(error);
            }
            catch (const exception &e)
            {
                result = NBErrors::SetLast(e);
            }
            catch (...)
            {
                result = NBErrors::SetLast(NBErrors::NB_ERROR_EXTERNAL_CPP);
            }
            return result;
        }
    }
}

#endif // !NB_DEVICE_HPP
