#ifndef NB_DEVICES_HPP
#define NB_DEVICES_HPP

#include <NBDevice.hpp>
#include <NBDevicesTypes.hpp>

namespace NextBiometrics
{
    namespace Devices
    {
        #include <NBDevices.h>

        class NBDevicesDeviceChangedEventArgs : public EventArgs
        {
            friend class NBDevices;

            NBDevice m_device;

            NBDevicesDeviceChangedEventArgs(const NBDevice& device, void * pParam)
                : EventArgs(pParam), m_device(device)
            {
            }
        public:
            const NBDevice& GetDevice() const
            {
                return m_device;
            }
        };

        typedef void(*NBDevicesDeviceChangedEvent)(const NBDevicesDeviceChangedEventArgs &args);

        class NBDevices
        {
            class NBDevicesDeviceChangedEventHandler
            {
                friend class NBDevices;

                NBDevicesDeviceChangedEvent m_pAddedCallback;
                NBDevicesDeviceChangedEvent m_pRemovedCallback;
                void * m_pParam;

                NBResult Process(const NBDeviceInfoA * psDeviceInfo, NBBool bIsAdded) const
                {
                    NBResult result = NBErrors::NB_OK;
                    try
                    {
                        HNBDevice hDevice = nullptr;
                        NBErrors::RequireNoErr(NBDeviceConnectA(psDeviceInfo, 0, &hDevice));
                        const NBDevicesDeviceChangedEventArgs args(NBDevice(hDevice), this->m_pParam);
                        if (bIsAdded) 
                            this->m_pAddedCallback(args);
                        else
                            this->m_pRemovedCallback(args);
                    }
                    catch (const NBError &error)
                    {
                        result = NBErrors::SetLast(error);
                    }
                    catch (const exception &e)
                    {
                        result = NBErrors::SetLast(e);
                    }
                    catch (...)
                    {
                        result = NBErrors::SetLast(NBErrors::NB_ERROR_EXTERNAL_CPP);
                    }
                    return result;
                }

                NBDevicesDeviceChangedEventHandler(const NBDevicesDeviceChangedEvent pAddedCallback, const NBDevicesDeviceChangedEvent pRemovedCallback, void * pParam)
                    : m_pAddedCallback(pAddedCallback), m_pRemovedCallback(pRemovedCallback), m_pParam(pParam)
                {
                }
            public:
                static NBResult NB_API AddedNativeCallback(const NBDeviceInfoA * psDeviceInfo, void * pParam)
                {
                    NBDevicesDeviceChangedEventHandler * pHandler = reinterpret_cast<NBDevicesDeviceChangedEventHandler *>(pParam);
                    return pHandler->Process(psDeviceInfo, NBTrue);
                }

                static NBResult NB_API RemovedNativeCallback(const NBDeviceInfoA * psDeviceInfo, void * pParam)
                {
                    NBDevicesDeviceChangedEventHandler * pHandler = reinterpret_cast<NBDevicesDeviceChangedEventHandler *>(pParam);
                    return pHandler->Process(psDeviceInfo, NBFalse);
                }
            };
        public:
            static bool IsInitialized()
            {
                NBBool bIsInitialized;
                NBErrors::RequireNoErr(NBDevicesIsInitialized(&bIsInitialized));
                return bIsInitialized == NBTrue ? true : false;
            }

            static void Initialize()
            {
                Initialize(nullptr, nullptr, nullptr);
            }

            static void Initialize(const NBDevicesDeviceChangedEvent pDeviceAddedCallback, const NBDevicesDeviceChangedEvent pDeviceRemovedCallback, void * pParam)
            {
                Initialize(pDeviceAddedCallback, pDeviceRemovedCallback, pParam, 0);
            }

            static void Initialize(const NBDevicesDeviceChangedEvent pDeviceAddedCallback, const NBDevicesDeviceChangedEvent pDeviceRemovedCallback, void * pParam, NBUInt uiFlags)
            {
                if (!IsInitialized())
                {
                    static unique_ptr<NBDevicesDeviceChangedEventHandler> m_deviceChangedEventHandler;
                    m_deviceChangedEventHandler.reset(new NBDevicesDeviceChangedEventHandler(pDeviceAddedCallback, pDeviceRemovedCallback, pParam));
                    NBErrors::RequireNoErr(NBDevicesInitializeA(pDeviceAddedCallback ? NBDevicesDeviceChangedEventHandler::AddedNativeCallback : nullptr, pDeviceRemovedCallback ? NBDevicesDeviceChangedEventHandler::RemovedNativeCallback : nullptr,
                        m_deviceChangedEventHandler.get(), uiFlags));
                }
            }

            static void Terminate()
            {
                NBErrors::RequireNoErr(NBDevicesTerminate());
            }

            static vector<NBDevice> GetDevices()
            {
                NBUInt uiDeviceCount;
                NBErrors::RequireNoErr(NBDevicesGetDevicesA(nullptr, 0, &uiDeviceCount));
                vector<NBDevice> arDevices(uiDeviceCount);
                if (uiDeviceCount > 0)
                {
                    vector<NBDeviceInfoA> arsDeviceInfos(uiDeviceCount);
                    NBUInt uiNewDeviceCount;
                    NBErrors::RequireNoErr(NBDevicesGetDevicesA(arsDeviceInfos.data(), NBUInt(arsDeviceInfos.size()), &uiNewDeviceCount));
                    for (NBUInt i = 0; i < min(uiNewDeviceCount, uiDeviceCount); i++)
                    {
                        HNBDevice hDevice = nullptr;
                        NBErrors::RequireNoErr(NBDeviceConnectA(&arsDeviceInfos[i], 0, &hDevice));
                        arDevices[i] = NBDevice(hDevice);
                    }
                }
                return arDevices;
            }
        };
    }
}

#endif // !NB_DEVICES_HPP
