#define _CRT_SECURE_NO_WARNINGS
// Platform headers
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// SDK headers
#include <NBDevicesLibrary.h>
#include <NBBiometricsLibrary.h>

#define SPI_DEFAULT_SPI_NAME        "/dev/spidev0.0"

//Default SYSFS path to access GPIO pins 
#define SPI_DEFAULT_SYSFS_PATH      "/sys/class/gpio"
/* 
   In case of SYSFS provide GPIO pin numbers. 
   To run from console
   execuite the command "./ NBDevicesConsoleSampleC - spi /dev/spidev0.0 459 402 408"
*/
NBInt SPI_DEFAULT_AWAKE_PIN       = 459;
NBInt SPI_DEFAULT_RESET_PIN       = 402;
NBInt SPI_DEFAULT_CHIP_SELECT_PIN = 408; 

#define MAKE_PIN(CHIP_NUMBER, LINE_NUMBER) (0x01000000 | (CHIP_NUMBER<<16) | LINE_NUMBER)        
/* 
   In case of /dev/gpiochipX interface, uncomment the below lines and
   pass the chip number and pin number to MAKE_PIN macro. 
   To run from console
   execuite the command "./ NBDevicesConsoleSampleC - spi /dev/spidev0.0 0 69 0 12 0 18" */   
/*
NBInt SPI_DEFAULT_AWAKE_PIN         =  MAKE_PIN(0, 69);
NBInt SPI_DEFAULT_RESET_PIN         =  MAKE_PIN(0, 12);
NBInt SPI_DEFAULT_CHIP_SELECT_PIN   =  MAKE_PIN(0, 18); 
*/

// Spoof Detect Fake, Live Score and Default threshold
#define NB_DEFAULT_FINGER_SPOOF 0
#define NB_DEFAULT_FINGER_SPOOF_LIVE 1000
#define NB_DEFAULT_FINGER_SPOOF_THRESHOLD 363

static NBResult PrintError(NBResult error, const NBAChar * szFunction)
{
    if (NBFailed(error))
    {
        NBUInt uiValueLength;
        NBAChar * szMessage;
        if (NBFailed(NBErrorsGetMessageA(error, NULL, 0, &uiValueLength)))
        {
            printf("ERROR: Failed to get error message length, original error: %d\n", error);
            return error;
        }
        szMessage = (NBAChar*)malloc(sizeof(NBAChar) * (uiValueLength + 1));
        if (!szMessage)
        {
            printf("ERROR: Failed to allocate memory for error message\n");
            return error;
        }
        if (NBFailed(NBErrorsGetMessageA(error, szMessage, uiValueLength + 1, &uiValueLength)))
        {
            printf("ERROR: Failed to get error message, original error: %d\n", error);
            free(szMessage);
            return error;
        }
        printf("ERROR: Function %s failed, error code: %d, error message: %s\n", szFunction, error, szMessage);
        free(szMessage);
        return error;
    }
    return error;
}

#define CAL_DATA_FILEPATH_SIZE              (NB_DEVICE_STRING_MAX_LENGTH + 50)

// CAK (customer assigned key) to be used for SDK Demo purposes
char g_DefaultCAKIdentifier[] = "DefaultCAKKey1";
uint8_t g_DefaultCAK[] = {
    0x05, 0x4B, 0x38, 0x3A, 0xCF, 0x5B, 0xB8, 0x01, 0xDC, 0xBB, 0x85, 0xB4, 0x47, 0xFF, 0xF0, 0x79, 
    0x77, 0x90, 0x90, 0x81, 0x51, 0x42, 0xC1, 0xBF, 0xF6, 0xD1, 0x66, 0x65, 0x0A, 0x66, 0x34, 0x11
};
// CDK (customer defined key) to be used for SDK Demo purposes
uint8_t g_DefaultCDK[] = {
    0x6B, 0xC5, 0x51, 0xD1, 0x12, 0xF7, 0xE3, 0x42, 0xBD, 0xDC, 0xFB, 0x5D, 0x79, 0x4E, 0x5A, 0xD6,
    0x54, 0xD1, 0xC9, 0x90, 0x28, 0x05, 0xCF, 0x5E, 0x4C, 0x83, 0x63, 0xFB, 0xC2, 0x3C, 0xF6, 0xAB
};
// This is the authentication key AUTH1, one of two keys which allow to establish a secure session
// Below is its initial/default value. The owner of the device can change it to his private secret value.
// For improved security the key should be obfuscated and not stored in plaintext as in this simple example.
static uint8_t defaultAuthKey1[] = {
    0xDA, 0x2E, 0x35, 0xB6, 0xCB, 0x96, 0x2B, 0x5F, 0x9F, 0x34, 0x1F, 0xD1, 0x47, 0x41, 0xA0, 0x4D,
    0xA4, 0x09, 0xCE, 0xE8, 0x35, 0x48, 0x3C, 0x60, 0xFB, 0x13, 0x91, 0xE0, 0x9E, 0x95, 0xB2, 0x7F
};

/*
    Summary:
     Load calibration data to the device from an input binary file.
    Parameters:
     hDevice :      [in] handle to device
    Remarks:
     The function loads the calibration data only for 65210-S type device, other devices are ignored and the calibration data won't be loaded.
     The input file shall be located in ./NBData directory and its name has to be ./<Device serial number>_calblob.bin
    Returns:
     If the function succeeds, the return value is NB_OK. If the
     function fails, one of the error codes is returned (<link Error codes>).
*/
static NBResult loadCalDataToDevice(const HNBDevice hDevice)
{
    NBAChar szFilepath[CAL_DATA_FILEPATH_SIZE];
    NBUInt uiSerialNoLength;
    NBResult tResult;
    FILE* pfTargetFile = NULL;
    NBByte* pCalDataBuffer = NULL;
    NBSizeType stCalDataSize;
    NBSizeType stSizeCheck;

    // ---- Calibration data file path construction ----
    // Format: ./<Device serial number>_calblob.bin

    tResult = NBDeviceGetSerialNumberA(hDevice, szFilepath, NB_DEVICE_STRING_MAX_LENGTH, &uiSerialNoLength);
    if (NBFailed(tResult))
    {
        PrintError(tResult, "NBDeviceGetSerialNumberA");
        goto finally;
    }

    sprintf(&szFilepath[uiSerialNoLength], "_calblob.bin");

    // ---- Calibration data file readout and update in the device ----
    pfTargetFile = fopen(szFilepath, "rb");
    if (NULL == pfTargetFile)
    {
        tResult = NB_ERROR_FAILED;
        printf("ERROR: Failed to open the calibration data file. \nThe input file path and name shall be in format: \"./<Device serial number>_calblob.bin\". \nFor this device specifically: \"%s\".\n \nNote: Use the NBBiometricsCreateEnrollTemplateConsoleSampleC sample with parameter -calibrate to generate data\n",
                szFilepath);
        goto finally;
    }

    fseek(pfTargetFile, 0L, SEEK_END);
    stCalDataSize = ftell(pfTargetFile);

    rewind(pfTargetFile);

    pCalDataBuffer = (NBByte*)malloc(stCalDataSize);
    if (NULL == pCalDataBuffer)
    {
        tResult = NB_ERROR_FAILED;
        printf("ERROR: loadCalDataToDevice: Memory allocation failed.\n");
        goto finally;
    }

    stSizeCheck = fread(pCalDataBuffer, 1, stCalDataSize, pfTargetFile);
    if (stSizeCheck != stCalDataSize)
    {
        tResult = NB_ERROR_FAILED;
        printf("ERROR: loadCalDataToDevice: failure in the calibration data file readout.\n");
        goto finally;
    }

    tResult = NBDeviceSetBlobParameter(hDevice, NB_DEVICE_BLOB_PARAMETER_CALIBRATION_DATA, pCalDataBuffer, stCalDataSize);
    if (NBFailed(tResult))
    {
        PrintError(tResult, "NBDeviceSetBlobParameter");
        goto finally;
    }

finally:
    if (pCalDataBuffer) free(pCalDataBuffer);
    if (pfTargetFile) fclose(pfTargetFile);
    return tResult;
}

static NBResult InitDevice(HNBDevice hDevice)
{
    NBResult result;
    NBDeviceCapabilities *psCapabilities = NULL;
    printf("Getting capabilities ...\n");
    result = NBDeviceGetCapabilities(hDevice, &psCapabilities);
    if (NBFailed(result))
    {
        PrintError(result, "NBDevicesGetCapabilities");
        return result;
    }

    switch (psCapabilities->eSecurityModel)
    {
    case NBDeviceSecurityModelNone:
        break;
    case NBDeviceSecurityModel65100:
        result = NBDeviceOpenSession(hDevice, NBDEVICE_65100_AUTH1_ID, sizeof(NBDEVICE_65100_AUTH1_ID), defaultAuthKey1, sizeof(defaultAuthKey1));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceOpenSession");
            return result;
        }
        break;
    case NBDeviceSecurityModel65200CakOnly:
        result = NBDeviceOpenSession(hDevice, g_DefaultCAKIdentifier, sizeof(g_DefaultCAKIdentifier), g_DefaultCAK, sizeof(g_DefaultCAK));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceOpenSession");
            return result;
        }
        break;
    case NBDeviceSecurityModel65200CakCdk:
        // Open secure session with the CDK
        result = NBDeviceOpenSession(hDevice, NBDEVICE_65200_CDK_IDENTIFIER, sizeof(NBDEVICE_65200_CDK_IDENTIFIER), g_DefaultCDK, sizeof(g_DefaultCDK));
        if (!NBFailed(result))
        {
            // This section shows the reset of CDK in the device
            // Delete CDK
            result = NBDeviceSetBlobParameter(hDevice, NB_DEVICE_BLOB_PARAMETER_SET_CDK, NULL, 0);
            if (NBFailed(result))
            {
                PrintError(result, "NBDeviceSetBlobParameter");
                return result;
            }
            // Close CDK session
            result = NBDeviceCloseSession(hDevice);
            if (NBFailed(result))
            {
                PrintError(result, "NBDeviceDisconnect");
                return result;
            }
        }
        // Open secure session with the CAK
        result = NBDeviceOpenSession(hDevice, g_DefaultCAKIdentifier, sizeof(g_DefaultCAKIdentifier), g_DefaultCAK, sizeof(g_DefaultCAK));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceOpenSession");
            return result;
        }
        // Set CDK
        result = NBDeviceSetBlobParameter(hDevice, NB_DEVICE_BLOB_PARAMETER_SET_CDK, g_DefaultCDK, sizeof(g_DefaultCDK));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceSetBlobParameter");
            return result;
        }
        // Close CAK session
        result = NBDeviceCloseSession(hDevice);
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceDisconnect");
            return result;
        }
        // Open secure session with the CDK
        result = NBDeviceOpenSession(hDevice, NBDEVICE_65200_CDK_IDENTIFIER, sizeof(NBDEVICE_65200_CDK_IDENTIFIER), g_DefaultCDK, sizeof(g_DefaultCDK));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceOpenSession");
            return result;
        }
        break;
    default:
        PrintError(result = NB_ERROR_ARGUMENT_OUT_OF_RANGE, "NBDevicesGetCapabilities");
        return result;
    }

    if (psCapabilities->bRequiresExternalCalibrationData)
    {
        result = loadCalDataToDevice(hDevice);
        if (NBFailed(result))
        {
            PrintError(result, "loadCalDataToDevice");
            return result;
        }
    }

    NBDeviceFree((NBByte*)psCapabilities);
    return result;
}

static NBResult PrintDeviceInformationA(const HNBDevice hDevice, NBDeviceScanFormat * peScanFormat)
{
    NBUInt uiActualLength;
    NBDeviceType type;
    NBDeviceConnectionType connectionType;
    NBAChar szTemp[NB_DEVICE_STRING_MAX_LENGTH];
    NBResult result;
    NBVersion sVersion;
    NBUInt i, uiSupportedScanTypes;
    NBDeviceScanFormat * areScanFormats = NULL;

    printf("Getting device id ...\n");
    result = NBDeviceGetIdA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetIdA");
        goto finally;
    }
    printf("Device id: %s\n", szTemp);

    printf("Getting device manufacturer ...\n");
    result = NBDeviceGetManufacturerA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetManufacturerA");
        goto finally;
    }
    printf("Device manufacturer: %s\n", szTemp);

    printf("Getting device model ...\n");
    result = NBDeviceGetModelA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetModelA");
        goto finally;
    }
    printf("Device model: %s\n", szTemp);

    printf("Getting device serial number ...\n");
    result = NBDeviceGetSerialNumberA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetSerialNumberA");
        goto finally;
    }
    printf("Device serial number: %s\n", szTemp);

    printf("Getting device product ...\n");
    result = NBDeviceGetProductA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetProductA");
        goto finally;
    }
    printf("Device product: %s\n", szTemp);

    printf("Getting device type ...\n");
    result = NBDeviceGetType(hDevice, &type);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetType");
        goto finally;
    }
    printf("Device type: %d\n", type);

    printf("Getting device connection type ...\n");
    result = NBDeviceGetConnectionType(hDevice, &connectionType);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetConnectionType");
        goto finally;
    }
    printf("Device connection type: %d\n", connectionType);

    printf("Getting firmware version ...\n");
    result = NBDeviceGetFirmwareVersion(hDevice, &sVersion);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetFirmwareVersion");
        goto finally;
    }
    printf("Firmware version: %d.%d.%d.%d\n", sVersion.iMajor, sVersion.iMinor, sVersion.iBuild, sVersion.iRevision);

    printf("Getting supported scan formats ...\n");
    result = NBDeviceGetSupportedScanFormats(hDevice, NULL, 0, &uiSupportedScanTypes);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetSupportedScanFormats");
        goto finally;
    }
    areScanFormats = (NBDeviceScanFormat*)calloc(uiSupportedScanTypes, sizeof(NBDeviceScanFormat));
    if (!areScanFormats)
    {
        result = NB_ERROR_OUT_OF_MEMORY;
        printf("ERROR: Failed to allocate memory for formats\n");
        goto finally;
    }
    result = NBDeviceGetSupportedScanFormats(hDevice, areScanFormats, uiSupportedScanTypes, &uiSupportedScanTypes);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetSupportedScanFormats");
        goto finally;
    }
    for (i = 0; i < uiSupportedScanTypes; i++)
    {
        NBDeviceScanFormatInfo sScanFormatInfo;
        result = NBDeviceGetScanFormatInfo(hDevice, areScanFormats[i], &sScanFormatInfo);
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceGetScanFormatInfo");
            goto finally;
        }
        printf("\t%d) %d (%d), width: %d, height: %d, horizontal resolution: %d, vertical resolution: %d\n", i + 1, sScanFormatInfo.eScanFormat, sScanFormatInfo.eScanFormatType, sScanFormatInfo.uiWidth, sScanFormatInfo.uiHeight, sScanFormatInfo.uiHorizontalResolution, sScanFormatInfo.uiVerticalResolution);
    }
    printf("Selecting first scan format ...\n");
    *peScanFormat = areScanFormats[0];
    finally:
    if (areScanFormats) free(areScanFormats);
    return result;
}

typedef struct
{
    NBBool bUseSpi;

    const NBAChar * szSpiName;
    NBInt iAwakePin;
    NBInt iResetPin;
    NBInt iChipSelectPin;
} SampleOptions;

static NBResult ConnectToDeviceSpi(NBDeviceIOParams * psOptions, HNBDevice * phDevice)
{
    NBResult result;
    HNBDevice hDevice = NULL;

    psOptions->phDevice = &hDevice;
    psOptions->uiFlags = 0;

    if (psOptions->iChipSelectPin == -1)
    {
        printf("Connecting to device %s (awake pin: %d, reset pin: %d) ...\n", psOptions->szSpiName, psOptions->iAwakePin, psOptions->iResetPin);
        result = NBDeviceConnectToSpiA(psOptions);
    }
    else
    {
        printf("Connecting to device %s (awake pin: %d, reset pin: %d, chip select pin: %d) ...\n", psOptions->szSpiName, psOptions->iAwakePin, psOptions->iResetPin, psOptions->iChipSelectPin);
        result = NBDeviceConnectToSpiExA(psOptions);
    }
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceConnectToSpiA/NBDeviceConnectToSpiExA");
        return result;
    }
    *phDevice = *psOptions->phDevice;
    return NB_OK;
}

static NBResult ConnectToDeviceUsb(HNBDevice * phDevice)
{
    NBResult result;
    NBDeviceInfoA arsDeviceInfos[NB_DEVICE_MAX_COUNT];
    NBUInt i, uiDeviceCount = 0;
    HNBDevice hDevice = NULL;

    printf("Enumerating devices ...\n");
    result = NBDevicesGetDevicesA(arsDeviceInfos, NB_DEVICE_MAX_COUNT, &uiDeviceCount);
    if (NBFailed(result))
    {
        PrintError(result, "NBDevicesGetDevicesA");
        return result;
    }

    if (uiDeviceCount == 0)
    {
        printf("No devices connected ...\n");
        return NB_ERROR_FAILED;
    }

    for (i = 0; i < uiDeviceCount; i++)
    {
        printf("%d)\t%s\n", arsDeviceInfos[i].uiDeviceIndex, arsDeviceInfos[i].szId);
        printf("\t%s %s: %s\n", arsDeviceInfos[i].szManufacturer, arsDeviceInfos[i].szModel, arsDeviceInfos[i].szSerialNumber);
    }

    printf("Connecting to device %d ...\n", arsDeviceInfos[0].uiDeviceIndex);
    result = NBDeviceConnectA(&arsDeviceInfos[0], 0, &hDevice);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceConnectA");
        return result;
    }

    printf("Connected ... \n");
    *phDevice = hDevice;
    return NB_OK;
}

static NBResult ParseCommandOptions(int argc, char *argv[], NBDeviceIOParams * psOptions)
{
    psOptions->bAntispoofing = NBFalse;
    psOptions->iSpoofThreshold = NB_DEFAULT_FINGER_SPOOF_THRESHOLD;
    psOptions->bUseSpi = NBFalse;
    strcpy(psOptions->szSpiName, SPI_DEFAULT_SPI_NAME);
    strcpy(psOptions->szSysfsPath, SPI_DEFAULT_SYSFS_PATH);
    psOptions->iAwakePin = SPI_DEFAULT_AWAKE_PIN;
    psOptions->iResetPin = SPI_DEFAULT_RESET_PIN;
    psOptions->iChipSelectPin = SPI_DEFAULT_CHIP_SELECT_PIN;

    for (int argindex = 0; argindex < (unsigned int)argc; argindex++)
    {
        if ((strcmp(argv[argindex], "-spoof") == 0))
        {
            psOptions->bAntispoofing = NBTrue;
            psOptions->iSpoofThreshold = NB_DEFAULT_FINGER_SPOOF_THRESHOLD;
        }
        if ((strcmp(argv[argindex], "-threshold") == 0))
        {
            if (argc > (argindex + 1))
            {
                if ((atoi(argv[argindex + 1])) > 0)
                {
                    argindex = argindex + 1;
                    psOptions->iSpoofThreshold = atoi(argv[argindex]);
                }
                else
                {
                    psOptions->iSpoofThreshold = NB_DEFAULT_FINGER_SPOOF_THRESHOLD;
                }
            }
            else
            {
                psOptions->iSpoofThreshold = NB_DEFAULT_FINGER_SPOOF_THRESHOLD;
            }
        }
        if (strcmp(argv[argindex], "-usb") == 0)
        {
            psOptions->bUseSpi = NBFalse;
            continue;
        }
        if (strcmp(argv[argindex], "-sysfspath") == 0)
        {
            if ((argindex + 1) >= argc)
            {
                printf("Not enough parameters for -sysfspath\n");
                return NB_ERROR_FAILED;
            }
            strcpy(psOptions->szSysfsPath, argv[argindex + 1]);
            argindex += 1;
            continue;
        }
        if (strcmp(argv[argindex], "-spi") == 0)
        {
            psOptions->bUseSpi = NBTrue;
            bool isSysInterface = false;
            bool isCharInterface = false;
            int j = 1;
            for (; j <= (argc - (argindex + 1)); ++j)
            {
                if (argv[argindex + j][0] == '-')
                {
                    break;
                }
            }
            if (j == 5)
            {
                isSysInterface = true;
            }
            else if (j == 8)
            {
                isCharInterface = true;
            }
            else
            {
                printf("Not enough parameters for SPI\n");
                return NB_ERROR_FAILED;
            }
            if (isCharInterface)
            {
                strcpy(psOptions->szSpiName, argv[argindex + 1]);
                psOptions->iAwakePin = MAKE_PIN(atoi(argv[argindex + 2]), atoi(argv[argindex + 3]));
                psOptions->iResetPin = MAKE_PIN(atoi(argv[argindex + 4]), atoi(argv[argindex + 5]));
                psOptions->iChipSelectPin = MAKE_PIN(atoi(argv[argindex + 6]), atoi(argv[argindex + 7]));
                argindex = argindex + 7;
                continue;
            }
            if (isSysInterface)
            {
                strcpy(psOptions->szSpiName, argv[argindex + 1]);
                psOptions->iAwakePin = atoi(argv[argindex + 2]);
                psOptions->iResetPin = atoi(argv[argindex + 3]);
                psOptions->iChipSelectPin = atoi(argv[argindex + 4]);
                argindex = argindex + 4;
                continue;
            }
        }
    }
    return NB_OK;
}

static NBResult ReadFileA(const NBAChar * szFileName, NBByte ** ppBuffer, NBSizeType * pstBufferSize)
{
    FILE * pFile = NULL;
    long lSize;

    pFile = fopen(szFileName, "rb");
    if (!pFile)
    {
        printf("ERROR: Failed to open file\n");
        return NB_ERROR_IO;
    }

    fseek(pFile, 0, SEEK_END);
    lSize = ftell(pFile);
    fseek(pFile, 0, SEEK_SET);  //same as rewind(f);

    *ppBuffer = (NBByte*)malloc(sizeof(NBByte) * lSize);
    if (!*ppBuffer)
    {
        fclose(pFile);
        printf("ERROR: Failed to allocate memory for contents of file\n");
        return NB_ERROR_OUT_OF_MEMORY;
    }

    if (fread(*ppBuffer, sizeof(NBByte), lSize, pFile) != lSize)
    {
        fclose(pFile);
        free(*ppBuffer); *ppBuffer = NULL;
        printf("ERROR: Failed to read contents of file\n");
        return NB_ERROR_IO;
    }
    *pstBufferSize = lSize;
    fclose(pFile);
    return NB_OK;
}

static const NBAChar * TemplateTypeToStrA(NBBiometricsTemplateType eTemplateType)
{
    switch (eTemplateType)
    {
    case NBBiometricsTemplateTypeProprietary:
        return "Proprietary";
    case NBBiometricsTemplateTypeIso:
        return "ISO";
    case NBBiometricsTemplateTypeAnsi:
        return "ANSI";
    default:
        return "Unknown NBBiometricsTemplateType value";
    }
}

static NBResult LoadTemplateA(const HNBBiometricsContext hContext, NBBiometricsTemplateType eTemplateType, const NBAChar * szFileName, NBVoid ** ppTemplate, NBSizeType * pstTemplateSize)
{
    NBResult result;
    NBSizeType stMaxSize;
    NBVoid * pTemplate = NULL;
    NBSizeType stTemplateSize;
    NBByte * pBuffer = NULL;
    NBSizeType stBufferSize;

    printf("Reading template from file %s (template type: %s)...\n", szFileName, TemplateTypeToStrA(eTemplateType));
    result = ReadFileA(szFileName, &pBuffer, &stBufferSize);
    if (NBFailed(result))
    {
        goto finally;
    }

    // Get max template size
    result = NBBiometricsContextGetMaxTemplateSize(hContext, eTemplateType, &stMaxSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextGetMaxTemplateSize");
        goto finally;
    }

    // Allocate buffer
    pTemplate = (NBVoid *)malloc(sizeof(NBByte) * stMaxSize);
    if (!pTemplate)
    {
        result = NB_ERROR_OUT_OF_MEMORY;
        printf("ERROR: Failed to allocate memory for template\n");
        goto finally;
    }

    result = NBBiometricsContextLoadTemplateFromMemory(hContext, eTemplateType, pBuffer, stBufferSize,
        0, pTemplate, stMaxSize, &stTemplateSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextLoadTemplateFromMemory");
        goto finally;
    }
    printf("Template loaded successfully.\n");
    *ppTemplate = pTemplate; pTemplate = NULL;
    *pstTemplateSize = stTemplateSize;
finally:
    if (pBuffer) free(pBuffer);
    if (pTemplate) free(pTemplate);
    return result;
}

static NBResult SaveTemplateA(const HNBBiometricsContext hContext, const NBAChar * szFileName, const NBVoid * pTemplate, NBSizeType stTemplateSize)
{
    NBByte * pBuffer = NULL;
    NBSizeType stBufferSize, stMaxSize;
    NBResult result;
    NBBiometricsTemplateType eTemplateType;
    FILE * pFile = NULL;

    result = NBBiometricsTemplateGetType(pTemplate, stTemplateSize, &eTemplateType);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsTemplateGetType");
        goto finally;
    }

    result = NBBiometricsContextGetMaxTemplateSize(hContext, eTemplateType, &stMaxSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextGetMaxTemplateSize");
        goto finally;
    }

    pBuffer = (NBByte*)malloc(sizeof(NBByte) * stMaxSize);
    if (!pBuffer)
    {
        result = NB_ERROR_OUT_OF_MEMORY;
        printf("ERROR: Failed to allocate memory for save buffer\n");
        goto finally;
    }

    result = NBBiometricsContextSaveTemplateToMemory(hContext, pTemplate, stTemplateSize, 0, pBuffer, stMaxSize, &stBufferSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextSaveTemplateToMemory");
        goto finally;
    }

    pFile = fopen(szFileName, "wb");
    if (!pFile)
    {
        result = NB_ERROR_FAILED;
        printf("ERROR: Failed to open file for writing\n");
        goto finally;
    }

    // Write Buffer
    if (fwrite(pBuffer, sizeof(NBByte), stBufferSize, pFile) != stBufferSize)
    {
        result = NB_ERROR_FAILED;
        printf("ERROR: Failed to write template to file\n");
    }

    // Close File
finally:
    if (pBuffer) free(pBuffer);
    if (pFile) fclose(pFile);
    return result;
}

static NBResult PrintUsage(int argc, char *argv[])
{
    printf("Usage:\n");
    printf("%s <source template format> <source template filename> <destination template format> <destination template filename> [<interface>]\n", argv[0]);
    printf("\t<source template format> - format of source template (proprietary, iso, ansi, isocc)\n");
    printf("\t<source template filename> - filename/path of source template\n");
    printf("\t<destination template format> - format of destination template (proprietary, iso, ansi, isocc)\n");
    printf("\t<destination template filename> - filename/path where to save destination template\n");
    printf("Interface:\n");
    printf("\t-usb - sample works in usb mode\n");
    printf("\t-spi - sample works in spi mode and tries to connect to specified SPI\n");
    printf("\tadditional spi parameters (optional): <spi name> <awake pin> <reset pin> <chip select pin>\n");
    printf("\t\t<spi name> - spi to connect to (default: %s)\n", SPI_DEFAULT_SPI_NAME);
    printf("\t\t<awake pin> - awake pin number to be used (default: %d)\n", SPI_DEFAULT_AWAKE_PIN);
    printf("\t\t<reset pin> - reset pin number to be used (default: %d)\n", SPI_DEFAULT_RESET_PIN);
    printf("\t\t<chip select pin> - chip select pin number to be used (default: %d, specify \"-1\" if not available)\n", SPI_DEFAULT_CHIP_SELECT_PIN);
    printf("\t-sysfspath - SYSFS path for SPI GPIO pins(awake, reset and chip select), default path: /sys/class/gpio \n");
    return NB_ERROR_FAILED;
}

static NBBiometricsTemplateType ParseTemplateType(const NBAChar * szStr)
{
    NBBiometricsTemplateType eTemplateType = NBBiometricsTemplateTypeUnknown;
    if (strcmp(szStr, "iso") == 0)
        eTemplateType = NBBiometricsTemplateTypeIso;
    else if (strcmp(szStr, "isocc") == 0)
        eTemplateType = NBBiometricsTemplateTypeIsoCompactCard;
    else if (strcmp(szStr, "ansi") == 0)
        eTemplateType = NBBiometricsTemplateTypeAnsi;
    else if (strcmp(szStr, "proprietary") == 0)
        eTemplateType = NBBiometricsTemplateTypeProprietary;
    return eTemplateType;
}

int main(int argc, char *argv[])
{
    NBResult result;
    NBVersion sVersion;
    HNBDevice hDevice = NULL;
    NBDeviceScanFormat eScanFormat;
    NBBool bTerminate = NBFalse;
    HNBBiometricsContext hContext = NULL;
    NBBool bIsSupported = NBFalse;
    NBBiometricsAlgorithmInfo sAlgorithmInfo;
    NBBiometricsTemplateType eSourceTemplateType;
    const NBAChar * szSourceTemplateFileName;
    NBBiometricsTemplateType eDestinationTemplateType;
    const NBAChar * szDestinationTemplateFileName;
    NBVoid * pSourceTemplate = NULL;
    NBSizeType stSourceTemplateSize = 0;
    NBVoid * pDestinationTemplate = NULL;
    NBSizeType stMaxTemplateSize;
    NBSizeType stDestinationTemplateSize;
    NBDeviceIOParams sOptions = { 0 };
    
    printf("%s\n", argv[0]);
#ifdef NB_WINDOWS
    printf("Sample application might require administrative privileges. If sensor is connected, but not detected - try restarting application in administrator mode (Run As Administrator)\n");
#endif
    if (argc < 5)
    {
        return PrintUsage(argc, argv);
    }

    eSourceTemplateType = ParseTemplateType(argv[1]);
    if (eSourceTemplateType == NBBiometricsTemplateTypeUnknown)
    {
        return PrintUsage(argc, argv);
    }
    szSourceTemplateFileName = argv[2];

    eDestinationTemplateType = ParseTemplateType(argv[3]);
    if (eDestinationTemplateType == NBBiometricsTemplateTypeUnknown)
    {
        return PrintUsage(argc, argv);
    }
    szDestinationTemplateFileName = argv[4];

    result = ParseCommandOptions(argc, &argv[0], &sOptions);
    if (NBFailed(result))
    {
        PrintError(result, "ParseCommandOptions Failed.");
        goto finally;
    }
    printf("Sample is in %s mode\n", sOptions.bUseSpi ? "SPI" : "USB");

    printf("Getting biometrics library version ...\n");
    result = NBBiometricsLibraryGetVersion(&sVersion);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsLibraryGetVersion");
        goto finally;
    }
    printf("Biometrics library version: %d.%d.%d.%d\n", sVersion.iMajor, sVersion.iMinor, sVersion.iBuild, sVersion.iRevision);

    printf("Getting devices library version ...\n");
    result = NBDevicesLibraryGetVersion(&sVersion);
    if (NBFailed(result))
    {
        PrintError(result, "NBDevicesLibraryGetVersion");
        goto finally;
    }
    printf("Devices library version: %d.%d.%d.%d\n", sVersion.iMajor, sVersion.iMinor, sVersion.iBuild, sVersion.iRevision);

    printf("Initializing devices ...\n");
    result = NBDevicesInitializeA(NULL, NULL, NULL, 0);
    if (NBFailed(result))
    {
        PrintError(result, "NBDevicesInitializeA");
        goto finally;
    }
    bTerminate = NBTrue;

    if (sOptions.bUseSpi)
        result = ConnectToDeviceSpi(&sOptions, &hDevice);
    else
        result = ConnectToDeviceUsb(&hDevice);
    if (NBFailed(result))
    {
        goto finally;
    }
    result = InitDevice(hDevice);
    if (NBFailed(result))
    {
        goto finally;
    }
    result = PrintDeviceInformationA(hDevice, &eScanFormat);
    if (NBFailed(result))
    {
        goto finally;
    }

    printf("Creating biometrics context ...\n");
    result = NBBiometricsContextCreate(hDevice, 0, &hContext);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextCreate");
        goto finally;
    }

    printf("Get biometrics algorithm info ...\n");
    result = NBBiometricsContextGetAlgorithmInfo(hContext, &sAlgorithmInfo);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextGetAlgorithmInfo");
        goto finally;
    }
    printf("Algorithm info: vendor id: %d, version: %d.%d.%d.%d\n", sAlgorithmInfo.uiId, sAlgorithmInfo.sVersion.iMajor, 
        sAlgorithmInfo.sVersion.iMinor, sAlgorithmInfo.sVersion.iBuild, sAlgorithmInfo.sVersion.iRevision);

    result = NBBiometricsContextIsTemplateTypeSupported(hContext, eSourceTemplateType, &bIsSupported);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextIsTemplateTypeSupported");
        goto finally;
    }
    if (!bIsSupported)
    {
        result = NB_ERROR_NOT_SUPPORTED;
        printf("ERROR: Source template type is not supported\n");
        goto finally;
    }

    result = NBBiometricsContextIsTemplateTypeSupported(hContext, eDestinationTemplateType, &bIsSupported);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextIsTemplateTypeConvertSupported");
        goto finally;
    }
    if (!bIsSupported)
    {
        result = NB_ERROR_NOT_SUPPORTED;
        printf("ERROR: Destination convertion template type is not supported\n");
        goto finally;
    }

    printf("Loading source template ...\n");
    result = LoadTemplateA(hContext, eSourceTemplateType, szSourceTemplateFileName, &pSourceTemplate, &stSourceTemplateSize);
    if (NBFailed(result))
    {
        goto finally;
    }

    result = NBBiometricsContextGetMaxTemplateSize(hContext, eDestinationTemplateType, &stMaxTemplateSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextGetMaxTemplateSize");
        goto finally;
    }

    printf("Allocating destination template ...\n");
    pDestinationTemplate = (NBByte*)malloc(stMaxTemplateSize * sizeof(NBByte));
    if (!pDestinationTemplate)
    {
        result = NB_ERROR_OUT_OF_MEMORY;
        printf("ERROR: Failed to allocate memory for template\n");
        goto finally;
    }

    printf("Converting template ...\n");
    result = NBBiometricsContextConvertTemplate(hContext, pSourceTemplate, stSourceTemplateSize, eDestinationTemplateType, 0, 
        pDestinationTemplate, stMaxTemplateSize, &stDestinationTemplateSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextConvertTemplate");
        goto finally;
    }

    printf("Saving destination template ...\n");
    result = SaveTemplateA(hContext, szDestinationTemplateFileName, pDestinationTemplate, stDestinationTemplateSize);
    if (NBFailed(result))
    {
        goto finally;
    }
    printf("Finished\n");
finally:
    if (hContext)
    {
        NBResult tResult = NBBiometricsContextDestroy(hContext);
        if (NBFailed(tResult))
        {
            PrintError(tResult, "NBBiometricsContextDestroy");
        }
    }
    if (hDevice)
    {
        NBResult tResult = NBDeviceDestroy(hDevice);
        if (NBFailed(tResult))
        {
            PrintError(tResult, "NBDeviceDestroy");
        }
    }
    if (bTerminate)
    {
        NBResult tResult = NBDevicesTerminate();
        if (NBFailed(tResult))
        {
            PrintError(tResult, "NBDevicesTerminate");
        }
    }
    if (pSourceTemplate) free(pSourceTemplate);
    if (pDestinationTemplate) free(pDestinationTemplate);

    return result;
}
