#define _CRT_SECURE_NO_WARNINGS
// Platform headers
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// SDK headers
#include <NBDevicesLibrary.h>
#include <NBBiometricsLibrary.h>
// Timestamp headers
#if !defined(NB_FreeBSD)
#include <sys/timeb.h>
#else
#include <sys/time.h>
#endif

#define SPI_DEFAULT_SPI_NAME        "/dev/spidev0.0"

//Default SYSFS path to access GPIO pins 
#define SPI_DEFAULT_SYSFS_PATH      "/sys/class/gpio"
/* 
   In case of SYSFS provide GPIO pin numbers. 
   To run from console
   execuite the command "./ NBDevicesConsoleSampleC - spi /dev/spidev0.0 459 402 408"
*/
NBInt SPI_DEFAULT_AWAKE_PIN       = 459;
NBInt SPI_DEFAULT_RESET_PIN       = 402;
NBInt SPI_DEFAULT_CHIP_SELECT_PIN = 408; 

#define MAKE_PIN(CHIP_NUMBER, LINE_NUMBER) (0x01000000 | (CHIP_NUMBER<<16) | LINE_NUMBER)        
/* 
   In case of /dev/gpiochipX interface, uncomment the below lines and
   pass the chip number and pin number to MAKE_PIN macro. 
   To run from console
   execuite the command "./ NBDevicesConsoleSampleC - spi /dev/spidev0.0 0 69 0 12 0 18" */   
/*
NBInt SPI_DEFAULT_AWAKE_PIN         =  MAKE_PIN(0, 69);
NBInt SPI_DEFAULT_RESET_PIN         =  MAKE_PIN(0, 12);
NBInt SPI_DEFAULT_CHIP_SELECT_PIN   =  MAKE_PIN(0, 18); 
*/	

#define CAL_DATA_FILEPATH_SIZE              (NB_DEVICE_STRING_MAX_LENGTH + 50)
#define CAL_DATA_FILEPATH_FILENAME_OFFSET   9

// Spoof Detect Fake, Live Score and Default threshold
#define NB_DEFAULT_FINGER_SPOOF 0
#define NB_DEFAULT_FINGER_SPOOF_LIVE 1000
#define NB_DEFAULT_FINGER_SPOOF_THRESHOLD 363

// CAK (customer assigned key) to be used for SDK Demo purposes, used for 65200
char g_DefaultCAKIdentifier[] = "DefaultCAKKey1";
uint8_t g_DefaultCAK[] = {
    0x05, 0x4B, 0x38, 0x3A, 0xCF, 0x5B, 0xB8, 0x01, 0xDC, 0xBB, 0x85, 0xB4, 0x47, 0xFF, 0xF0, 0x79, 
    0x77, 0x90, 0x90, 0x81, 0x51, 0x42, 0xC1, 0xBF, 0xF6, 0xD1, 0x66, 0x65, 0x0A, 0x66, 0x34, 0x11
};
// CDK (customer defined key) to be used for SDK Demo purposes, used for 65200
uint8_t g_DefaultCDK[] = {
    0x6B, 0xC5, 0x51, 0xD1, 0x12, 0xF7, 0xE3, 0x42, 0xBD, 0xDC, 0xFB, 0x5D, 0x79, 0x4E, 0x5A, 0xD6,
    0x54, 0xD1, 0xC9, 0x90, 0x28, 0x05, 0xCF, 0x5E, 0x4C, 0x83, 0x63, 0xFB, 0xC2, 0x3C, 0xF6, 0xAB
};
// This is the authentication key AUTH1, for 65100/65110, one of two keys which allow to establish a secure session
// Below is its initial/default value. The owner of the device can change it to his private secret value.
// For improved security the key should be obfuscated and not stored in plaintext as in this simple example.
static uint8_t defaultAuthKey1[] = {
    0xDA, 0x2E, 0x35, 0xB6, 0xCB, 0x96, 0x2B, 0x5F, 0x9F, 0x34, 0x1F, 0xD1, 0x47, 0x41, 0xA0, 0x4D,
    0xA4, 0x09, 0xCE, 0xE8, 0x35, 0x48, 0x3C, 0x60, 0xFB, 0x13, 0x91, 0xE0, 0x9E, 0x95, 0xB2, 0x7F
};

/*
  Threshold for spoof assessment - if the image score is bigger than this value then is a live scan.
  Available values: 0 (spoof) - 1000 (live)). Set the value to -1 to reset to default.
*/
static int iSpoofScore = NB_DEFAULT_FINGER_SPOOF;

/*
 * Total Time consumption calculation for Scan and Extract Operation.
*/
static bool isFingerPresent = false;
int diff_time_millis;
#if !defined(NB_FreeBSD)
struct timeb wall_start, wall_end;
#else
struct timeval wall_start, wall_end;
#endif
static NBResult PrintError(NBResult error, const NBAChar * szFunction)
{
    if (NBFailed(error))
    {
        NBUInt uiValueLength;
        NBAChar * szMessage;
        if (NBFailed(NBErrorsGetMessageA(error, NULL, 0, &uiValueLength)))
        {
            printf("ERROR: Failed to get error message length, original error: %d\n", error);
            return error;
        }
        szMessage = (NBAChar*)malloc(sizeof(NBAChar) * (uiValueLength + 1));
        if (!szMessage)
        {
            printf("ERROR: Failed to allocate memory for error message\n");
            return error;
        }
        if (NBFailed(NBErrorsGetMessageA(error, szMessage, uiValueLength + 1, &uiValueLength)))
        {
            printf("ERROR: Failed to get error message, original error: %d\n", error);
            free(szMessage);
            return error;
        }
        printf("ERROR: Function %s failed, error code: %d, error message: %s\n", szFunction, error, szMessage);
        free(szMessage);
        return error;
    }
    return error;
}

/*
    Summary:
     Load calibration data to the device from an input binary file.
    Parameters:
     hDevice :      [in] handle to device
    Remarks:
     The function loads the calibration data only for 65210-S type device, other devices are ignored and the calibration data won't be loaded.
     The input file shall be located in ./NBData directory and its name has to be ./<Device serial number>_calblob.bin
    Returns:
     If the function succeeds, the return value is NB_OK. If the
     function fails, one of the error codes is returned (<link Error codes>).
*/
static NBResult loadCalDataToDevice(const HNBDevice hDevice)
{
    NBAChar szFilepath[CAL_DATA_FILEPATH_SIZE];
    NBUInt uiSerialNoLength;
    NBResult tResult;
    FILE* pfTargetFile = NULL;
    NBByte* pCalDataBuffer = NULL;
    NBSizeType stCalDataSize;
    NBSizeType stSizeCheck;
    
    // ---- Calibration data file path construction ----
    // Format: ./<Device serial number>_calblob.bin

    tResult = NBDeviceGetSerialNumberA(hDevice, szFilepath, NB_DEVICE_STRING_MAX_LENGTH, &uiSerialNoLength);
    if (NBFailed(tResult))
    {
        PrintError(tResult, "NBDeviceGetSerialNumberA");
        goto finally;
    }

    sprintf(&szFilepath[uiSerialNoLength], "_calblob.bin");

    // ---- Calibration data file readout and update in the device ----
    pfTargetFile = fopen(szFilepath, "rb");
    if (NULL == pfTargetFile)
    {
        tResult = NB_ERROR_FAILED;
        printf("ERROR: Failed to open the calibration data file. \nThe input file path and name shall be in format: \"./<Device serial number>_calblob.bin\". \nFor this device specifically: \"%s\".\n \nNote: Use the NBBiometricsCreateEnrollTemplateConsoleSampleC sample with parameter -calibrate to generate data\n",
            szFilepath);
        goto finally;
    }

    fseek(pfTargetFile, 0L, SEEK_END);
    stCalDataSize = ftell(pfTargetFile);

    rewind(pfTargetFile);

    pCalDataBuffer = (NBByte*)malloc(stCalDataSize);
    if (NULL == pCalDataBuffer)
    {
        tResult = NB_ERROR_FAILED;
        printf("ERROR: loadCalDataToDevice: Memory allocation failed.\n");
        goto finally;
    }

    stSizeCheck = fread(pCalDataBuffer, 1, stCalDataSize, pfTargetFile);
    if (stSizeCheck != stCalDataSize)
    {
        tResult = NB_ERROR_FAILED;
        printf("ERROR: loadCalDataToDevice: failure in the calibration data file readout.\n");
        goto finally;
    }

    tResult = NBDeviceSetBlobParameter(hDevice, NB_DEVICE_BLOB_PARAMETER_CALIBRATION_DATA, pCalDataBuffer, stCalDataSize);
    if (NBFailed(tResult))
    {
        PrintError(tResult, "NBDeviceSetBlobParameter");
        goto finally;
    }

    finally:
    if (pCalDataBuffer) free(pCalDataBuffer);
    if (pfTargetFile) fclose(pfTargetFile);
    return tResult;
}

static NBResult InitDevice(HNBDevice hDevice)
{
    NBResult result;
    NBDeviceCapabilities *psCapabilities = NULL;
    printf("Getting capabilities ...\n");
    result = NBDeviceGetCapabilities(hDevice, &psCapabilities);
    if (NBFailed(result))
    {
        PrintError(result, "NBDevicesGetCapabilities");
        return result;
    }

    switch (psCapabilities->eSecurityModel)
    {
    case NBDeviceSecurityModelNone:
        break;
    case NBDeviceSecurityModel65100:
        result = NBDeviceOpenSession(hDevice, NBDEVICE_65100_AUTH1_ID, sizeof(NBDEVICE_65100_AUTH1_ID), defaultAuthKey1, sizeof(defaultAuthKey1));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceOpenSession");
            return result;
        }
        break;
    case NBDeviceSecurityModel65200CakOnly:
        result = NBDeviceOpenSession(hDevice, g_DefaultCAKIdentifier, sizeof(g_DefaultCAKIdentifier), g_DefaultCAK, sizeof(g_DefaultCAK));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceOpenSession");
            return result;
        }
        break;
    case NBDeviceSecurityModel65200CakCdk:
        // Open secure session with the CDK
        result = NBDeviceOpenSession(hDevice, NBDEVICE_65200_CDK_IDENTIFIER, sizeof(NBDEVICE_65200_CDK_IDENTIFIER), g_DefaultCDK, sizeof(g_DefaultCDK));
        if (!NBFailed(result))
        {
            // This section shows the reset of CDK in the device
            // Delete CDK
            result = NBDeviceSetBlobParameter(hDevice, NB_DEVICE_BLOB_PARAMETER_SET_CDK, NULL, 0);
            if (NBFailed(result))
            {
                PrintError(result, "NBDeviceSetBlobParameter");
                return result;
            }
            // Close CDK session
            result = NBDeviceCloseSession(hDevice);
            if (NBFailed(result))
            {
                PrintError(result, "NBDeviceDisconnect");
                return result;
            }
        }
        // Open secure session with the CAK
        result = NBDeviceOpenSession(hDevice, g_DefaultCAKIdentifier, sizeof(g_DefaultCAKIdentifier), g_DefaultCAK, sizeof(g_DefaultCAK));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceOpenSession");
            return result;
        }
        // Set CDK
        result = NBDeviceSetBlobParameter(hDevice, NB_DEVICE_BLOB_PARAMETER_SET_CDK, g_DefaultCDK, sizeof(g_DefaultCDK));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceSetBlobParameter");
            return result;
        }
        // Close CAK session
        result = NBDeviceCloseSession(hDevice);
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceDisconnect");
            return result;
        }
        // Open secure session with the CDK
        result = NBDeviceOpenSession(hDevice, NBDEVICE_65200_CDK_IDENTIFIER, sizeof(NBDEVICE_65200_CDK_IDENTIFIER), g_DefaultCDK, sizeof(g_DefaultCDK));
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceOpenSession");
            return result;
        }
        break;
    default:
        PrintError(result = NB_ERROR_ARGUMENT_OUT_OF_RANGE, "NBDevicesGetCapabilities");
        return result;
    }

    if (psCapabilities->bRequiresExternalCalibrationData)
    {
        result = loadCalDataToDevice(hDevice);
        if (NBFailed(result))
        {
            return result;
        }
    }

    NBDeviceFree((NBByte*)psCapabilities);
    return result;
}

static NBResult PrintDeviceInformationA(const HNBDevice hDevice, NBDeviceScanFormat * peScanFormat)
{
    NBUInt uiActualLength;
    NBDeviceType type;
    NBDeviceConnectionType connectionType;
    NBAChar szTemp[NB_DEVICE_STRING_MAX_LENGTH];
    NBResult result;
    NBVersion sVersion;
    NBUInt i, uiSupportedScanTypes;
    NBDeviceScanFormat * areScanFormats = NULL;

    printf("Getting device id ...\n");
    result = NBDeviceGetIdA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetIdA");
        goto finally;
    }
    printf("Device id: %s\n", szTemp);

    printf("Getting device manufacturer ...\n");
    result = NBDeviceGetManufacturerA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetManufacturerA");
        goto finally;
    }
    printf("Device manufacturer: %s\n", szTemp);

    printf("Getting device model ...\n");
    result = NBDeviceGetModelA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetModelA");
        goto finally;
    }
    printf("Device model: %s\n", szTemp);

    printf("Getting device serial number ...\n");
    result = NBDeviceGetSerialNumberA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetSerialNumberA");
        goto finally;
    }
    printf("Device serial number: %s\n", szTemp);

    printf("Getting device product ...\n");
    result = NBDeviceGetProductA(hDevice, szTemp, NB_DEVICE_STRING_MAX_LENGTH, &uiActualLength);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetProductA");
        goto finally;
    }
    printf("Device product: %s\n", szTemp);

    printf("Getting device type ...\n");
    result = NBDeviceGetType(hDevice, &type);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetType");
        goto finally;
    }
    printf("Device type: %d\n", type);

    printf("Getting device connection type ...\n");
    result = NBDeviceGetConnectionType(hDevice, &connectionType);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetConnectionType");
        goto finally;
    }
    printf("Device connection type: %d\n", connectionType);

    printf("Getting firmware version ...\n");
    result = NBDeviceGetFirmwareVersion(hDevice, &sVersion);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetFirmwareVersion");
        goto finally;
    }
    printf("Firmware version: %d.%d.%d.%d\n", sVersion.iMajor, sVersion.iMinor, sVersion.iBuild, sVersion.iRevision);

    printf("Getting supported scan formats ...\n");
    result = NBDeviceGetSupportedScanFormats(hDevice, NULL, 0, &uiSupportedScanTypes);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetSupportedScanFormats");
        goto finally;
    }
    areScanFormats = (NBDeviceScanFormat*)calloc(uiSupportedScanTypes, sizeof(NBDeviceScanFormat));
    if (!areScanFormats)
    {
        result = NB_ERROR_OUT_OF_MEMORY;
        printf("ERROR: Failed to allocate memory for formats\n");
        goto finally;
    }
    result = NBDeviceGetSupportedScanFormats(hDevice, areScanFormats, uiSupportedScanTypes, &uiSupportedScanTypes);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetSupportedScanFormats");
        goto finally;
    }
    for (i = 0; i < uiSupportedScanTypes; i++)
    {
        NBDeviceScanFormatInfo sScanFormatInfo;
        result = NBDeviceGetScanFormatInfo(hDevice, areScanFormats[i], &sScanFormatInfo);
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceGetScanFormatInfo");
            goto finally;
        }
        printf("\t%d) %d (%d), width: %d, height: %d, horizontal resolution: %d, vertical resolution: %d\n", i + 1, sScanFormatInfo.eScanFormat, sScanFormatInfo.eScanFormatType, sScanFormatInfo.uiWidth, sScanFormatInfo.uiHeight, sScanFormatInfo.uiHorizontalResolution, sScanFormatInfo.uiVerticalResolution);
    }
    printf("Selecting first scan format ...\n");
    *peScanFormat = areScanFormats[0];
    finally:
    if (areScanFormats) free(areScanFormats);
    return result;
}

static const NBAChar * BiometricsStatusToStrA(NBBiometricsStatus eStatus)
{
    switch (eStatus)
    {
    case NBBiometricsStatusOk: return "OK";
    case NBBiometricsStatusNone: return "None";
    case NBBiometricsStatusCanceled: return "Canceled";
    case NBBiometricsStatusTimeout: return "Timeout";
    case NBBiometricsStatusBadQuality: return "Bad quality";
    case NBBiometricsStatusTooFewMinutiae: return "Too few minutiae";
    case NBBiometricsStatusMatchNotFound: return "Match not found";
    case NBBiometricsStatusLatentDetected: return "Latent detected";
    case NBBiometricsStatusNeedMoreSamples: return "Need more samples";
    case NBBiometricsStatusSpoofDetected: return "Spoof Detected";
    default: return "Unknown NBBiometricsStatus value";
    }
}

static const NBAChar * ScanStatusToStrA(NBDeviceScanStatus eScanStatus)
{
    switch (eScanStatus)
    {
    case NBDeviceScanStatusNone: return "None";
    case NBDeviceScanStatusOk: return "OK";
    case NBDeviceScanStatusNoFinger: return "No finger";
    case NBDeviceScanStatusCanceled: return "Canceled";
    case NBDeviceScanStatusTimeout: return "Timeout";
    case NBDeviceScanStatusNotRemoved: return "Finger not removed";
    case NBDeviceScanStatusBadQuality: return "Bad quality";
    case NBDeviceScanStatusBadSize: return "Bad size";
    case NBDeviceScanStatusSpoof: return "Spoof";
    case NBDeviceScanStatusEmpty: return "Empty";
    case NBDeviceScanStatusDone: return "Done";
    case NBDeviceScanStatusLiftFinger: return "Lift Finger";
    case NBDeviceScanStatusWaitForSensorInitialization: return "Wait for Initialization";
    case NBDeviceScanStatusPutFingerOnSensor: return "Put Finger on Sensor";
    case NBDeviceScanStatusKeepFingerOnSensor: return "Keep Finger on Sensor";
    case NBDeviceScanStatusWaitForDataProcessing: return "Wait for Processing";
    case NBDeviceScanStatusSpoofDetected: return "Spoof Detected";
    default: return "Unknown NBDeviceScanStatus value";
    }
}

// NBBiometricsExtractConsoleSampleC to save the last successful extracted image in bmp format
static NBResult SaveBmp(const NBDeviceScanFormatInfo ** psScanFormatInfo, const NBUInt8 * pBuffer, NBSizeType stBufferSize, const NBAChar * szFileName)
{
    NBUInt uiPadLength = (4 - ((*psScanFormatInfo)->uiWidth) % 4) % 4;
    NBSizeType stFileSize = 54 + 4 * 256 + stBufferSize + uiPadLength * (*psScanFormatInfo)->uiHeight, stImageSize = stBufferSize + (uiPadLength * (*psScanFormatInfo)->uiHeight);

    NBUInt8 bmpfileheader[14] = { 'B','M', 0,0,0,0, 0,0, 0,0, 54,4,0,0 };
    NBUInt8 bmpinfoheader[40] = { 40,0,0,0,  // Number of bytes in the DIB header (from this point) 
                                  0,0,0,0,   // Width of the bitmap in pixels 
                                  0,0,0,0,   // Height of the bitmap in pixels.
                                  1,0,       // Number of color planes being used 
                                  8,0,       // Number of bits per pixel
                                  0,0,0,0,   // BI_RGB, no pixel array compression used 
                                  16,0,0,0,  // Size of the raw bitmap data(including padding)
                                  53,59,0,0, // Horizontal resolution of the image
                                  53,59,0,0, // Vertical resolution of the image
                                  0,1,0,0,   // Number of colors in the palette 
                                  0,0,0,0 }; // 0 means all colors are important 
    NBUInt8 bmppad[4] = { 0, 0, 0, 0 };
    NBUInt uiValue;

    bmpfileheader[2] = (NBUInt8)(stFileSize);
    bmpfileheader[3] = (NBUInt8)(stFileSize >> 8);
    bmpfileheader[4] = (NBUInt8)(stFileSize >> 16);
    bmpfileheader[5] = (NBUInt8)(stFileSize >> 24);

    bmpinfoheader[4] = (NBUInt8)((*psScanFormatInfo)->uiWidth);
    bmpinfoheader[5] = (NBUInt8)((*psScanFormatInfo)->uiWidth >> 8);
    bmpinfoheader[6] = (NBUInt8)((*psScanFormatInfo)->uiWidth >> 16);
    bmpinfoheader[7] = (NBUInt8)((*psScanFormatInfo)->uiWidth >> 24);
    bmpinfoheader[8] = (NBUInt8)((*psScanFormatInfo)->uiHeight);
    bmpinfoheader[9] = (NBUInt8)((*psScanFormatInfo)->uiHeight >> 8);
    bmpinfoheader[10] = (NBUInt8)((*psScanFormatInfo)->uiHeight >> 16);
    bmpinfoheader[11] = (NBUInt8)((*psScanFormatInfo)->uiHeight >> 24);

    bmpinfoheader[20] = (NBUInt8)(stImageSize);
    bmpinfoheader[21] = (NBUInt8)(stImageSize >> 8);
    bmpinfoheader[22] = (NBUInt8)(stImageSize >> 16);
    bmpinfoheader[23] = (NBUInt8)(stImageSize >> 24);

    uiValue = ((*psScanFormatInfo)->uiHorizontalResolution * 393701 + 5000)/10000;
    bmpinfoheader[24] = (NBUInt8)(uiValue);
    bmpinfoheader[25] = (NBUInt8)(uiValue >> 8);
    bmpinfoheader[26] = (NBUInt8)(uiValue >> 16);
    bmpinfoheader[27] = (NBUInt8)(uiValue >> 24);
    uiValue = ((*psScanFormatInfo)->uiVerticalResolution * 393701 + 5000)/10000;
    bmpinfoheader[28] = (NBUInt8)(uiValue);
    bmpinfoheader[29] = (NBUInt8)(uiValue >> 8);
    bmpinfoheader[30] = (NBUInt8)(uiValue >> 16);
    bmpinfoheader[31] = (NBUInt8)(uiValue >> 24);

    NBInt i;
    NBUInt8 colormask[4] = { 0, 0, 0, 255 };
    FILE * f = fopen(szFileName, "wb");
    if (f == NULL)
    {
        printf("Failed to open file: %s\n", szFileName);
        return NB_ERROR_FAILED;
    }

    fwrite(bmpfileheader, 1, 14, f);
    fwrite(bmpinfoheader, 1, 40, f);

    for (i = 0; i < 256; i++)
    {
        colormask[0] = i;
        colormask[1] = i;
        colormask[2] = i;
        fwrite(colormask, 4, 1, f);
    }

    for (i = ((*psScanFormatInfo)->uiHeight - 1); i >= 0; i--)
    {
        fwrite(pBuffer + ((*psScanFormatInfo)->uiWidth*i), 1, (*psScanFormatInfo)->uiWidth, f);
        fwrite(bmppad, 1, uiPadLength, f);
    }
    fclose(f);
    return NB_OK;
}

static NBResult NB_API ScanPreview(const HNBBiometricsContext hContext, NBBiometricsStatus eBiometricsStatus, const HNBDevice hDevice, NBDeviceScanStatus * peStatus, const NBDeviceScanFormatInfo * psScanFormatInfo, const NBByte * pBuffer, NBSizeType stBufferSize, const NBDeviceScanPreviewDetails * psDetails, void * pParam)
{
    NBResult result;
    if (psDetails->iFingerDetectValue >= NB_DEVICE_DEFAULT_FINGER_DETECT_THRESHOLD && !isFingerPresent)
    {
        //Start scan Time calculated here..    
        isFingerPresent = true;
        printf("Clock Started!! \n");
        #if !defined(NB_FreeBSD)
        ftime(&wall_start);
        printf("Start Time: %u.%03u (In Seconds) \n", wall_start.time, wall_start.millitm);
        #else
        gettimeofday(&wall_start,NULL);
        printf("Start Time: %ld.%ld (In Seconds) \n", wall_start.tv_sec, (wall_start.tv_usec/1000));
        #endif
    }
    printf("\tPreview: biometrics status: %s, scan status: %s, finger detect value: %d\n", BiometricsStatusToStrA(eBiometricsStatus), ScanStatusToStrA(*peStatus), psDetails ? psDetails->iFingerDetectValue : 0);
    if (eBiometricsStatus == NBBiometricsStatusNeedMoreSamples)
    {
        printf("\tPlease lift your fingerprint and press again (new sample requested)\n");
    }
    if ((psDetails != NULL) && (psDetails->iSpoofScore > NB_DEFAULT_FINGER_SPOOF) && (psDetails->iSpoofScore < NB_DEFAULT_FINGER_SPOOF_LIVE))
    {
        iSpoofScore = psDetails->iSpoofScore;
    }
    
    //Image captured successfully, saving to scan.bmp...
    result = SaveBmp(&psScanFormatInfo, pBuffer, stBufferSize, "scan.bmp");
    if (NBFailed(result))
    {
        PrintError(result, "Failed to saveBmp file");
    }
    return NB_OK;
}

typedef struct
{
    NBBool bUseSpi;
    const NBAChar * szSpiName;
    NBInt iAwakePin;
    NBInt iResetPin;
    NBInt iChipSelectPin;
    NBBool bAntispoofing;
    NBInt iSpoofThreshold;
} SampleOptions;

static NBResult ConnectToDeviceSpi(NBDeviceIOParams * psOptions, HNBDevice * phDevice)
{
    NBResult result;
    HNBDevice hDevice = NULL;

    psOptions->phDevice = &hDevice;
    psOptions->uiFlags = 0;

    if (psOptions->iChipSelectPin == -1)
    {
        printf("Connecting to device %s (awake pin: %d, reset pin: %d) ...\n", psOptions->szSpiName, psOptions->iAwakePin, psOptions->iResetPin);
        result = NBDeviceConnectToSpiA(psOptions);
    }
    else
    {
        printf("Connecting to device %s (awake pin: %d, reset pin: %d, chip select pin: %d) ...\n", psOptions->szSpiName, psOptions->iAwakePin, psOptions->iResetPin, psOptions->iChipSelectPin);
        result = NBDeviceConnectToSpiExA(psOptions);
    }
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceConnectToSpiA/NBDeviceConnectToSpiExA");
        return result;
    }
    *phDevice = *psOptions->phDevice;
    return NB_OK;
}

static NBResult ConnectToDeviceUsb(HNBDevice * phDevice)
{
    NBResult result;
    NBDeviceInfoA arsDeviceInfos[NB_DEVICE_MAX_COUNT];
    NBUInt i, uiDeviceCount = 0;
    HNBDevice hDevice = NULL;

    printf("Enumerating devices ...\n");
    result = NBDevicesGetDevicesA(arsDeviceInfos, NB_DEVICE_MAX_COUNT, &uiDeviceCount);
    if (NBFailed(result))
    {
        PrintError(result, "NBDevicesGetDevicesA");
        return result;
    }

    if (uiDeviceCount == 0)
    {
        printf("No devices connected ...\n");
        return NB_ERROR_FAILED;
    }

    for (i = 0; i < uiDeviceCount; i++)
    {
        printf("%d)\t%s\n", arsDeviceInfos[i].uiDeviceIndex, arsDeviceInfos[i].szId);
        printf("\t%s %s: %s\n", arsDeviceInfos[i].szManufacturer, arsDeviceInfos[i].szModel, arsDeviceInfos[i].szSerialNumber);
    }

    printf("Connecting to device %d ...\n", arsDeviceInfos[0].uiDeviceIndex);
    result = NBDeviceConnectA(&arsDeviceInfos[0], 0, &hDevice);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceConnectA");
        return result;
    }

    printf("Connected ... \n");
    *phDevice = hDevice;
    return NB_OK;
}

static NBResult ParseCommandOptions(int argc, char *argv[], NBDeviceIOParams * psOptions)
{
    psOptions->bAntispoofing = NBFalse;
    psOptions->iSpoofThreshold = NB_DEFAULT_FINGER_SPOOF_THRESHOLD;
    psOptions->bUseSpi = NBFalse;
    strcpy(psOptions->szSpiName, SPI_DEFAULT_SPI_NAME);
    strcpy(psOptions->szSysfsPath, SPI_DEFAULT_SYSFS_PATH);
    psOptions->iAwakePin = SPI_DEFAULT_AWAKE_PIN;
    psOptions->iResetPin = SPI_DEFAULT_RESET_PIN;
    psOptions->iChipSelectPin = SPI_DEFAULT_CHIP_SELECT_PIN;

    for (int argindex = 0; argindex < (unsigned int)argc; argindex++)
    {
        if ((strcmp(argv[argindex], "-spoof") == 0))
        {
            psOptions->bAntispoofing = NBTrue;
            psOptions->iSpoofThreshold = NB_DEFAULT_FINGER_SPOOF_THRESHOLD;
        }
        if ((strcmp(argv[argindex], "-threshold") == 0))
        {
            if (argc > (argindex + 1))
            {
                if ((atoi(argv[argindex + 1])) > 0)
                {
                    argindex = argindex + 1;
                    psOptions->iSpoofThreshold = atoi(argv[argindex]);
                }
                else
                {
                    psOptions->iSpoofThreshold = NB_DEFAULT_FINGER_SPOOF_THRESHOLD;
                }
            }
            else
            {
                psOptions->iSpoofThreshold = NB_DEFAULT_FINGER_SPOOF_THRESHOLD;
            }
        }
        if (strcmp(argv[argindex], "-usb") == 0)
        {
            psOptions->bUseSpi = NBFalse;
            continue;
        }
        if (strcmp(argv[argindex], "-sysfspath") == 0)
        {
            if ((argindex + 1) >= argc)
            {
                printf("Not enough parameters for -sysfspath\n");
                return NB_ERROR_FAILED;
            }
            strcpy(psOptions->szSysfsPath, argv[argindex + 1]);
            argindex += 1;
            continue;
        }
        if (strcmp(argv[argindex], "-spi") == 0)
        {
            psOptions->bUseSpi = NBTrue;
            bool isSysInterface = false;
            bool isCharInterface = false;
            int j = 1;
            for (; j <= (argc - (argindex + 1)); ++j)
            {
                if (argv[argindex + j][0] == '-')
                {
                    break;
                }
            }
            if (j == 5)
            {
                isSysInterface = true;
            }
            else if (j == 8)
            {
                isCharInterface = true;
            }
            else
            {
                printf("Not enough parameters for SPI\n");
                return NB_ERROR_FAILED;
            }
            if (isCharInterface)
            {
                strcpy(psOptions->szSpiName, argv[argindex + 1]);
                psOptions->iAwakePin = MAKE_PIN(atoi(argv[argindex + 2]), atoi(argv[argindex + 3]));
                psOptions->iResetPin = MAKE_PIN(atoi(argv[argindex + 4]), atoi(argv[argindex + 5]));
                psOptions->iChipSelectPin = MAKE_PIN(atoi(argv[argindex + 6]), atoi(argv[argindex + 7]));
                argindex = argindex + 7;
                continue;
            }
            if (isSysInterface)
            {
                strcpy(psOptions->szSpiName, argv[argindex + 1]);
                psOptions->iAwakePin = atoi(argv[argindex + 2]);
                psOptions->iResetPin = atoi(argv[argindex + 3]);
                psOptions->iChipSelectPin = atoi(argv[argindex + 4]);
                argindex = argindex + 4;
                continue;
            }
        }
    }
    return NB_OK;
}

static NBResult SaveTemplateA(const HNBBiometricsContext hContext, const NBAChar * szFileName, const NBVoid * pTemplate, NBSizeType stTemplateSize)
{
    NBByte * pBuffer = NULL;
    NBSizeType stBufferSize, stMaxSize;
    NBResult result;
    NBBiometricsTemplateType eTemplateType;
    FILE * pFile = NULL;

    result = NBBiometricsTemplateGetType(pTemplate, stTemplateSize, &eTemplateType);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsTemplateGetType");
        goto finally;
    }

    result = NBBiometricsContextGetMaxTemplateSize(hContext, eTemplateType, &stMaxSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextGetMaxTemplateSize");
        goto finally;
    }

    pBuffer = (NBByte*)malloc(sizeof(NBByte) * stMaxSize);
    if (!pBuffer)
    {
        result = NB_ERROR_OUT_OF_MEMORY;
        printf("ERROR: Failed to allocate memory for save buffer\n");
        goto finally;
    }

    result = NBBiometricsContextSaveTemplateToMemory(hContext, pTemplate, stTemplateSize, 0, pBuffer, stMaxSize, &stBufferSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextSaveTemplateToMemory");
        goto finally;
    }

    pFile = fopen(szFileName, "wb");
    if (!pFile)
    {
        result = NB_ERROR_FAILED;
        printf("ERROR: Failed to open file for writing\n");
        goto finally;
    }

    // Write Buffer
    if (fwrite(pBuffer, sizeof(NBByte), stBufferSize, pFile) != stBufferSize)
    {
        result = NB_ERROR_FAILED;
        printf("ERROR: Failed to write template to file\n");
    }

    // Close File
finally:
    if (pBuffer) free(pBuffer);
    if (pFile) fclose(pFile);
    return result;
}

static NBResult PrintUsage(int argc, char *argv[])
{
    printf("Usage:\n");
    printf("%s <template type> <template filename> [<interface>]\n", argv[0]);
    printf("\t<template type> - format of template to be extracted (proprietary, iso, ansi, isocc)\n");
    printf("\t<template filename> - filename to save extracted template to\n");
    printf("Interface:\n");
    printf("\t-usb - sample works in usb mode\n");
    printf("\t-spi - sample works in spi mode and tries to connect to specified SPI\n");
    printf("\tadditional spi parameters (optional): <spi name> <awake pin> <reset pin> <chip select pin>\n");
    printf("\t\t<spi name> - spi to connect to (default: %s)\n", SPI_DEFAULT_SPI_NAME);
    printf("\t\t<awake pin> - awake pin number to be used (default: %d)\n", SPI_DEFAULT_AWAKE_PIN);
    printf("\t\t<reset pin> - reset pin number to be used (default: %d)\n", SPI_DEFAULT_RESET_PIN);
    printf("\t\t<chip select pin> - chip select pin number to be used (default: %d, specify \"-1\" if not available)\n", SPI_DEFAULT_CHIP_SELECT_PIN);
    printf("\t-sysfspath - SYSFS path for SPI GPIO pins(awake, reset and chip select), default path: /sys/class/gpio \n");
    printf("\t-spoof - controls the antispoofing for modules that support it (enabled if parameter present, disabled if not)\n");
    printf("\t-threshold <threshold> - controls the antispoofing threshold\n");
    return NB_ERROR_FAILED;
}

static NBBiometricsTemplateType ParseTemplateType(const NBAChar * szStr)
{
    NBBiometricsTemplateType eTemplateType = NBBiometricsTemplateTypeUnknown;
    if (strcmp(szStr, "iso") == 0)
        eTemplateType = NBBiometricsTemplateTypeIso;
    else if (strcmp(szStr, "isocc") == 0)
        eTemplateType = NBBiometricsTemplateTypeIsoCompactCard;
    else if (strcmp(szStr, "ansi") == 0)
        eTemplateType = NBBiometricsTemplateTypeAnsi;
    else if (strcmp(szStr, "proprietary") == 0)
        eTemplateType = NBBiometricsTemplateTypeProprietary;
    return eTemplateType;
}

int main(int argc, char *argv[])
{
    NBResult result;
    NBVersion sVersion;
    HNBDevice hDevice = NULL;
    NBDeviceScanFormat eScanFormat;
    NBBool bTerminate = NBFalse;
    HNBBiometricsContext hContext = NULL;
    NBBiometricsAlgorithmInfo sAlgorithmInfo;
    NBBiometricsTemplateType eTemplateType;
    const NBAChar * szTemplateFileName;
    NBBiometricsFingerPosition eFingerPosition = NBBiometricsFingerPositionUnknown;
    NBVoid * pTemplate = NULL;
    NBBiometricsStatus eStatus;
    NBSizeType stActualTemplateSize;
    NBSizeType stMaxTemplateSize;
    NBBiometricsScanParams sScanParams;
    NBDeviceIOParams sOptions = { 0 };
    
    printf("%s\n", argv[0]);
#ifdef NB_WINDOWS
    printf("Sample application might require administrative privileges. If sensor is connected, but not detected - try restarting application in administrator mode (Run As Administrator)\n");
#endif
    if (argc < 3)
    {
        return PrintUsage(argc, argv);
    }

    eTemplateType = ParseTemplateType(argv[1]);
    if (eTemplateType == NBBiometricsTemplateTypeUnknown)
    {
        return PrintUsage(argc, argv);
    }
    szTemplateFileName = argv[2];

    result = ParseCommandOptions(argc, &argv[0], &sOptions);
    if (NBFailed(result))
    {
        PrintError(result, "ParseCommandOptions Failed.");
        goto finally;
    }
    printf("Sample is in %s mode\n", sOptions.bUseSpi ? "SPI" : "USB");

    printf("Getting biometrics library version ...\n");
    result = NBBiometricsLibraryGetVersion(&sVersion);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsLibraryGetVersion");
        goto finally;
    }
    printf("Biometrics library version: %d.%d.%d.%d\n", sVersion.iMajor, sVersion.iMinor, sVersion.iBuild, sVersion.iRevision);

    printf("Getting devices library version ...\n");
    result = NBDevicesLibraryGetVersion(&sVersion);
    if (NBFailed(result))
    {
        PrintError(result, "NBDevicesLibraryGetVersion");
        goto finally;
    }
    printf("Devices library version: %d.%d.%d.%d\n", sVersion.iMajor, sVersion.iMinor, sVersion.iBuild, sVersion.iRevision);

    printf("Initializing devices ...\n");
    result = NBDevicesInitializeA(NULL, NULL, NULL, 0);
    if (NBFailed(result))
    {
        PrintError(result, "NBDevicesInitializeA");
        goto finally;
    }
    bTerminate = NBTrue;

    if (sOptions.bUseSpi)
        result = ConnectToDeviceSpi(&sOptions, &hDevice);
    else
        result = ConnectToDeviceUsb(&hDevice);
    if (NBFailed(result))
    {
        goto finally;
    }
    result = InitDevice(hDevice);
    if (NBFailed(result))
    {
        goto finally;
    }
    if (sOptions.bAntispoofing)
    {
        result = NBDeviceSetParameter(hDevice, NB_DEVICE_PARAMETER_ANTISPOOF_ENABLED, sOptions.bAntispoofing);
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceSetParameter");
            goto finally;
        }
        result = NBDeviceSetParameter(hDevice, NB_DEVICE_PARAMETER_ANTISPOOF_THRESHOLD, sOptions.iSpoofThreshold);
        if (NBFailed(result))
        {
            PrintError(result, "NBDeviceSetParameter");
            goto finally;
        }
    }
    result = PrintDeviceInformationA(hDevice, &eScanFormat);
    if (NBFailed(result))
    {
        goto finally;
    }

    /*If we set to 1, the Latent detect is enabled and the one time captured BGImage is used further */
#if (0)
    NBDeviceScanFormatInfo sScanFormatInfo;
    NBUInt uiImageSize;
    NBByte * pBuffer = NULL;

    result = NBDeviceSetParameter(hDevice, NB_DEVICE_PARAMETER_SUBTRACT_BACKGROUND, NBTrue);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceSetParameter");
        goto finally;
    }

    result = NBDeviceGetScanFormatInfo(hDevice, eScanFormat, &sScanFormatInfo);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetScanFormatInfo");
        goto finally;
    }
    uiImageSize = sScanFormatInfo.uiHeight * sScanFormatInfo.uiWidth;
    pBuffer = (NBByte *)malloc(uiImageSize);
    result = NBDeviceScanBGImage(hDevice, eScanFormat, NULL, pBuffer, uiImageSize, &eStatus);
    if (NBFailed(result))
    {
        PrintError(result, "NBDeviceGetScanFormatInfo");
        goto finally;
    }
#endif

    printf("Creating biometrics context ...\n");
    result = NBBiometricsContextCreate(hDevice, 0, &hContext);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextCreate");
        goto finally;
    }

    printf("Get biometrics algorithm info ...\n");
    result = NBBiometricsContextGetAlgorithmInfo(hContext, &sAlgorithmInfo);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextGetAlgorithmInfo");
        goto finally;
    }
    printf("Algorithm info: vendor id: %d, version: %d.%d.%d.%d\n", sAlgorithmInfo.uiId, sAlgorithmInfo.sVersion.iMajor, 
        sAlgorithmInfo.sVersion.iMinor, sAlgorithmInfo.sVersion.iBuild, sAlgorithmInfo.sVersion.iRevision);

    result = NBBiometricsContextGetMaxTemplateSize(hContext, eTemplateType, &stMaxTemplateSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextGetMaxTemplateSize");
        goto finally;
    }

    printf("Allocating template ...\n");
    pTemplate = (NBByte*)malloc(stMaxTemplateSize * sizeof(NBByte));
    if (!pTemplate)
    {
        result = NB_ERROR_OUT_OF_MEMORY;
        printf("ERROR: Failed to allocate memory for template\n");
        goto finally;
    }

    sScanParams.eScanFormat = eScanFormat;
    sScanParams.iTimeout = NB_DEVICE_SCAN_TIMEOUT_INFINITE;
    sScanParams.pPreviewProc = ScanPreview;
    sScanParams.pParam = NULL;
    isFingerPresent = false;

    printf("Extraction started, please put your fingerprint on sensor!\n");
    iSpoofScore = NB_DEFAULT_FINGER_SPOOF;
    result = NBBiometricsContextExtractFromScan(hContext, eTemplateType, eFingerPosition, &sScanParams, 0, pTemplate, stMaxTemplateSize, &eStatus, &stActualTemplateSize);
    if (NBFailed(result))
    {
        PrintError(result, "NBBiometricsContextExtractFromScan");
        goto finally;
    }
    if ((iSpoofScore > NB_DEFAULT_FINGER_SPOOF) && (iSpoofScore < NB_DEFAULT_FINGER_SPOOF_LIVE))
    {
        printf("\tSpoof Score is: %d\n", iSpoofScore);
        if (iSpoofScore <= sOptions.iSpoofThreshold)
        {
            printf("Spoof Fingerprint Detected. Extraction failed.\n");
            goto finally;
        }
    }
    if (isFingerPresent)
    {
        //Caluculate End Time Here..
        #if !defined(NB_FreeBSD)
        ftime(&wall_end);
        diff_time_millis = (int)(1000.0 * (wall_end.time - wall_start.time) + (wall_end.millitm - wall_start.millitm));
        printf("End Time: %u.%03u (In Seconds) \n", wall_end.time, wall_end.millitm);
        #else
        gettimeofday(&wall_end,NULL);
        diff_time_millis = (int)(1000.0 * (wall_end.tv_sec - wall_start.tv_sec) + ((wall_end.tv_usec/1000) - (wall_start.tv_usec/1000)));
        printf("End Time: %u.%03u (In Seconds) \n", wall_end.tv_sec, (wall_end.tv_usec/1000));
        #endif
        printf("\nTotal time consumption:%fs \n", (float)diff_time_millis / 1000);
    }
    if (eStatus == NBBiometricsStatusOk)
    {
        printf("Extracted successfully. Saving template to %s ...\n", szTemplateFileName);
        result = SaveTemplateA(hContext, szTemplateFileName, pTemplate, stActualTemplateSize);
        if (NBFailed(result))
        {
            goto finally;
        }
    }
    else
    {
        printf("Failed to extract template, reason: %s\n", BiometricsStatusToStrA(eStatus));
    }
    printf("Finished\n");
finally:
    if (hContext)
    {
        NBResult tResult = NBBiometricsContextDestroy(hContext);
        if (NBFailed(tResult))
        {
            PrintError(tResult, "NBBiometricsContextDestroy");
        }
    }
    if (hDevice)
    {
        NBResult tResult = NBDeviceDestroy(hDevice);
        if (NBFailed(tResult))
        {
            PrintError(tResult, "NBDeviceDestroy");
        }
    }
    if (bTerminate)
    {
        NBResult tResult = NBDevicesTerminate();
        if (NBFailed(tResult))
        {
            PrintError(tResult, "NBDevicesTerminate");
        }
    }
    if (pTemplate) free(pTemplate);

    return result;
}
