﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using NBBiometricsSampleCS.Misc;
using NBBiometricsSampleCS.Tabs;
using NextBiometrics.Biometrics;
using NextBiometrics.Devices;
using System.Text;
using System.IO;
using System.ComponentModel;

namespace NBBiometricsSampleCS.Forms
{
    public partial class MainForm : Form
    {
        #region Private constants

        private const string SpiDefaultSpiName = "/dev/spidev0.0";
        private const int SpiDefaultAwakePin = 69;
        private const int SpiDefaultResetPin = 12;
        private const int SpiDefaultChipSelectPin = 18;
        private const uint BlobSetCDK = NextBiometrics.Devices.NBDevice.NB_DEVICE_BLOB_PARAMETER_SET_CDK;

        // CAK (customer assigned key) to be used for SDK Demo purposes
        private static string g_DefaultCAKIdentifier_string = "DefaultCAKKey1\0";
        private byte[] g_DefaultCAKIdentifier = Encoding.ASCII.GetBytes(g_DefaultCAKIdentifier_string);
        private byte[] g_DefaultCAK = {
            0x05, 0x4B, 0x38, 0x3A, 0xCF, 0x5B, 0xB8, 0x01, 0xDC, 0xBB, 0x85, 0xB4, 0x47, 0xFF, 0xF0, 0x79, 
            0x77, 0x90, 0x90, 0x81, 0x51, 0x42, 0xC1, 0xBF, 0xF6, 0xD1, 0x66, 0x65, 0x0A, 0x66, 0x34, 0x11
            };
        // CDK (customer defined key) to be used for SDK Demo purposes
        private static string NBU_CDK_IDENTIFIER_string = "Application Lock\0";
        private byte[] NBU_CDK_IDENTIFIER = Encoding.ASCII.GetBytes(NBU_CDK_IDENTIFIER_string);
        private byte[] g_DefaultCDK = {
            0x6B, 0xC5, 0x51, 0xD1, 0x12, 0xF7, 0xE3, 0x42, 0xBD, 0xDC, 0xFB, 0x5D, 0x79, 0x4E, 0x5A, 0xD6,
            0x54, 0xD1, 0xC9, 0x90, 0x28, 0x05, 0xCF, 0x5E, 0x4C, 0x83, 0x63, 0xFB, 0xC2, 0x3C, 0xF6, 0xAB
            };

        //NB65100U Support
        private static string NBDEVICE_65100_AUTH1_ID_string = "AUTH1\0";
        private byte[] NBDEVICE_65100_AUTH1_ID = Encoding.ASCII.GetBytes(NBDEVICE_65100_AUTH1_ID_string);

        private static string NBDEVICE_65100_AUTH2_ID_string = "AUTH2\0";
        private byte[] NBDEVICE_65100_AUTH2_ID = Encoding.ASCII.GetBytes(NBDEVICE_65100_AUTH2_ID_string);

        private byte[] defaultAuthKey1 = {
            0xDA, 0x2E, 0x35, 0xB6, 0xCB, 0x96, 0x2B, 0x5F, 0x9F, 0x34, 0x1F, 0xD1, 0x47, 0x41, 0xA0, 0x4D,
            0xA4, 0x09, 0xCE, 0xE8, 0x35, 0x48, 0x3C, 0x60, 0xFB, 0x13, 0x91, 0xE0, 0x9E, 0x95, 0xB2, 0x7F
        };

        #endregion

        #region Private fields

        private DeviceInfo _currentDevice;
        private NBBiometricsContext _currentContext;

        private Database _database = null;

        private readonly Dictionary<string, BaseTab> _tabs =
            new Dictionary<string, BaseTab>
            {
                { "Capture", new CaptureImageTab() },
                { "Extract", new ExtractTab() },
                { "Enroll", new EnrollTab() },
                { "Verify", new VerifyTab() },
                { "Identify", new IdentifyTab() },
                { "Database", new DatabaseTab() }
            };

        #endregion

        #region Public constructor

        public MainForm()
        {
            InitializeComponent();

            labelVersion.Text = NBDevicesLibrary.Version.ToString();
            labelBiometricVersion.Text = NBBiometricsLibrary.Version.ToString();

            foreach (var tab in _tabs)
            {
                TabPage newPage = new TabPage(tab.Key);
                newPage.UseVisualStyleBackColor = true;
                newPage.Controls.Add(tab.Value);
                tab.Value.Dock = DockStyle.Fill;
                tab.Value.BusyChanged += ValueOnBusyChanged;
                tabControlRight.TabPages.Add(newPage);
            }
        }

        #endregion

        #region Private form events

        private void MainForm_Load(object sender, EventArgs e)
        {
            buttonInitialize.PerformClick();
        }

        private void buttonInitialize_Click(object sender, EventArgs e)
        {
            if (NBDevices.IsInitialized) return;
            Initialize();
            buttonInitialize.Enabled = false;
            buttonTerminate.Enabled = true;
            buttonRefresh.Enabled = true;
            buttonConnectToSpi.Enabled = true;
            buttonDisconnect.Enabled = false;
            buttonRefresh.PerformClick();
        }

        private void buttonTerminate_Click(object sender, EventArgs e)
        {
            if (!NBDevices.IsInitialized) return;
            foreach (var tab in _tabs)
            {
                tab.Value.Cancel();
            }
            listBoxDevices.Items.Clear();
            if (_currentContext != null)
            {
                _currentContext.Dispose();
                _currentContext = null;
            }
            _currentDevice = null;
            ClearInfo();

            Terminate();
            buttonInitialize.Enabled = true;
            buttonTerminate.Enabled = false;
            buttonRefresh.Enabled = false;
            buttonConnectToSpi.Enabled = false;
            buttonDisconnect.Enabled = false;
        }

        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (_database != null) _database.Terminate();
            buttonTerminate.PerformClick();
        }

        private void buttonRefresh_Click(object sender, EventArgs e)
        {
            listBoxDevices.Items.Clear();
            if (_currentContext != null)
            {
                _currentContext.Dispose();
                _currentContext = null;
            }
            _currentDevice = null;
            ClearInfo();

            var devices = NBDevices.GetDevices().Select(x => new DeviceInfo(false, x)).ToArray();
            listBoxDevices.Items.AddRange(devices);
            if (devices.Length > 0)
            {
                listBoxDevices.SelectedIndex = 0;
            }
            else
            {
                // No device -- disable all tabs
                foreach (TabPage tab in tabControlRight.TabPages) tab.Enabled = false;
            }
        }

        private void listBoxDevices_SelectedValueChanged(object sender, EventArgs e)
        {
            _currentDevice = listBoxDevices.SelectedItem as DeviceInfo;
            if (_currentDevice == null)
            {
                _currentContext.Dispose();
                _currentContext = null;
                _database = null;
                ClearInfo();
            }
            else
            {
                UpdateInfo(_currentDevice);
                _currentContext = CreateContext(_currentDevice.Device);
                if (_currentContext != null && (HasExtraction(_currentContext) && (HasIdentification(_currentContext) || 
                    HasVerification(_currentContext))))
                {
                    if (_database == null)
                    {
                        _database = new Database(_currentContext);
                        _database.Initialize();
                    }
                    _database.BiometricsContext = _currentContext;
                }
                else
                {
                    _database = null;
                }
            }
            foreach (KeyValuePair<string, BaseTab> tab in _tabs)
            {
                tab.Value.Database = _database;
                tab.Value.BiometricsContext = _currentContext;
                tab.Value.Device = _currentDevice != null ? _currentDevice.Device : null;
            }
            foreach (TabPage tab in tabControlRight.TabPages)
            {
                if (tab.Controls[0] is CaptureImageTab)
                {
                    tab.Enabled = _currentDevice != null;
                }
                if (tab.Controls[0] is ExtractTab)
                {
                    tab.Enabled = _currentContext != null && HasExtraction(_currentContext);
                }
                if (tab.Controls[0] is VerifyTab)
                {
                    tab.Enabled = _currentContext != null && HasVerification(_currentContext);
                }
                if (tab.Controls[0] is IdentifyTab)
                {
                    tab.Enabled = _currentContext != null && HasIdentification(_currentContext);
                }
                if (tab.Controls[0] is EnrollTab || tab.Controls[0] is DatabaseTab)
                {
                    tab.Enabled = _currentContext != null && HasExtraction(_currentContext) &&
                                  (HasVerification(_currentContext) || HasIdentification(_currentContext));
                }
            }
        }

        private void buttonReset_Click(object sender, EventArgs e)
        {
            if (_currentDevice != null)
            {
                _currentDevice.Device.Reset();
                GetDeviceCapabilities(_currentDevice.Device);
            }
        }

        private void buttonDisconnect_Click(object sender, EventArgs e)
        {
            if (!_currentDevice.IsSpi) return;

            listBoxDevices.Items.Remove(_currentDevice);
        }

        private void buttonConnectToSpi_Click(object sender, EventArgs e)
        {
            using (var form = new ConnectToSpiForm())
            {
                form.SpiName = SpiDefaultSpiName;
                form.AwakePin = SpiDefaultAwakePin;
                form.ResetPin = SpiDefaultResetPin;
                form.ChipSelectPin = SpiDefaultChipSelectPin;

                if (form.ShowDialog() == DialogResult.OK)
                {
                    listBoxDevices.SelectedIndex = listBoxDevices.Items.Add(form.Device);
                }
            }
        }

        private void tabControlRight_SelectedIndexChanged(object sender, EventArgs e)
        {
            foreach (var tab in _tabs)
            {
                tab.Value.RefreshData();
            }
        }

        private void ValueOnBusyChanged(object sender, EventArgs e)
        {
            var isBusy = IsBusy();
            groupBoxDevices.Enabled = !isBusy;
            groupBoxDeviceInfo.Enabled = !isBusy;
        }

        #endregion

        #region Private methods

        private bool IsBusy()
        {
            bool isBusy = false;
            foreach (var tab in _tabs)
            {
                if (tab.Value.IsBusy)
                {
                    isBusy = true;
                    break;
                }
            }
            return isBusy;
        }

        private void Initialize()
        {
            NBDevices.Initialize(DeviceAddedEvent, DeviceRemovedEvent);
        }

        private void GetDeviceCapabilities(NBDevice device)
        {
            if (device != null)
            {
                DeviceCapabilities deviceCapabilities = device.GetCapabilities();

                if (deviceCapabilities.eSecurityModel != NBDeviceSecurityModel.ModelNone && device.IsSessionOpen())
                {
                    device.CloseSession();
                }
                switch (deviceCapabilities.eSecurityModel)
                {
                    case NBDeviceSecurityModel.ModelNone:
                        break;
                    case NBDeviceSecurityModel.Model65100:
                        device.OpenSession(NBDEVICE_65100_AUTH1_ID, defaultAuthKey1);
                        break;
                    case NBDeviceSecurityModel.Model65200CakOnly:
                        device.OpenSession(g_DefaultCAKIdentifier, g_DefaultCAK);
                        break;
                    case NBDeviceSecurityModel.Model65200CakCdk:
                        OpenCAKCDKSession(device);
                        break;
                    default:
                        throw new InvalidEnumArgumentException();
                }

                if (deviceCapabilities.bRequiresExternalCalibrationData)
                {
                    SetCalibrationBlob(device);
                }
            }
        }

        private void OpenCAKCDKSession(NBDevice device)
        {
            try
            {
                // Open secure session with the CDK
                device.OpenSession(NBU_CDK_IDENTIFIER, g_DefaultCDK);
                // This section shows the reset of CDK in the device
                // Delete CDK
                device.SetBlobParameter(BlobSetCDK, null);
                // Close CDK Session
                device.CloseSession();
            }
            catch { }
            // Open secure session with the CAK
            device.OpenSession(g_DefaultCAKIdentifier, g_DefaultCAK);
            // Set CDK
            device.SetBlobParameter(BlobSetCDK, g_DefaultCDK);
            // Close CAK session
            device.CloseSession();
            // Open secure session with the CDK
            device.OpenSession(NBU_CDK_IDENTIFIER, g_DefaultCDK);
        }

        private void OpenSecureSession(NBDevice device)
        {
            NBDeviceType tType = device.Type;

            if (tType == NBDeviceType.NB65100U)
            {
                // Open secure session with 65100U specific settings
                device.OpenSession(NBDEVICE_65100_AUTH1_ID, defaultAuthKey1);
            }
            else
            {
                // Open secure session with CAK
                device.OpenSession(g_DefaultCAKIdentifier, g_DefaultCAK);
            }
        }

        private void SetCalibrationBlob(NBDevice device)
        {
            string _fileName = Directory.GetCurrentDirectory() + device.SerialNumber + "_calblob.bin";
            try
            {
                byte[] _pBlob = File.ReadAllBytes(_fileName);
                device.SetBlobParameter(NBDevice.NB_DEVICE_BLOB_PARAMETER_CALIBRATION_DATA, _pBlob);
            }
            catch (DirectoryNotFoundException)
            {
                MessageBox.Show("Error: DirectoryNotFoundException - Calibration data file at " + _fileName + " cannot be found\nNote: Use the NBBiometricsCreateEnrollTemplateConsoleSampleC sample with parameter -calibrate to generate data");
                return;
            }
            catch (FileNotFoundException)
            {
                MessageBox.Show("Error: FileNotFoundException - Calibration data file at " + _fileName + " cannot be found\nNote: Use the NBBiometricsCreateEnrollTemplateConsoleSampleC sample with parameter -calibrate to generate data");
                return;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
                return;
            }
        }

        private void DeviceRemovedEvent(object sender, NBDevicesDeviceChangedEventArgs args)
        {
            if (InvokeRequired)
            {
                Invoke(new MethodInvoker(() => DeviceRemovedEvent(sender, args)));
                return;
            }
            listBoxDevices.Items.Remove(new DeviceInfo(false, args.Device));
        }

        private void DeviceAddedEvent(object sender, NBDevicesDeviceChangedEventArgs args)
        {
            if (InvokeRequired)
            {
                Invoke(new MethodInvoker(() => DeviceAddedEvent(sender, args)));
                return;
            }
            int index = listBoxDevices.Items.Add(new DeviceInfo(false, args.Device));
            if (!IsBusy())
                listBoxDevices.SelectedIndex = index;
        }

        private void Terminate()
        {
            if (NBDevices.IsInitialized)
            {
                NBDevices.Terminate();
            }
        }

        private void ClearInfo()
        {
            textBoxId.Text = textBoxManufacturer.Text = textBoxModel.Text = textBoxSerialNumber.Text = textBoxProduct.Text = string.Empty;
            textBoxFirmwareVersion.Text = textBoxType.Text = textBoxConnectionType.Text = textBoxSupportedFormats.Text = string.Empty;
            groupBoxDeviceInfo.Enabled = false;
            buttonDisconnect.Enabled = false;
        }

        private void UpdateInfo(DeviceInfo deviceInfo)
        {
            ClearInfo();
            var device = deviceInfo.Device;
            GetDeviceCapabilities(device);
            buttonDisconnect.Enabled = deviceInfo.IsSpi;
            textBoxId.Text = device.Id;
            textBoxManufacturer.Text = device.Manufacturer;
            textBoxModel.Text = device.Model;
            textBoxSerialNumber.Text = device.SerialNumber;
            textBoxProduct.Text = device.Product;
            textBoxFirmwareVersion.Text = device.FirmwareVersion.ToString();
            textBoxType.Text = device.Type.ToString();
            textBoxConnectionType.Text = device.ConnectionType.ToString();
            textBoxSupportedFormats.Text = string.Join(Environment.NewLine,
                device.SupportedScanFormats.Select(x => x.ToString()).ToArray());
            groupBoxDeviceInfo.Enabled = true;
        }

        private static bool HasIdentification(NBBiometricsContext context)
        {
            foreach (NBBiometricsTemplateTypeInfo templateTypeInfo in context.SupportedTemplateTypes)
            {
                if (templateTypeInfo.IsIdentificationSupported)
                {
                    return true;
                }
            }
            return false;
        }

        private static bool HasVerification(NBBiometricsContext context)
        {
            foreach (NBBiometricsTemplateTypeInfo templateTypeInfo in context.SupportedTemplateTypes)
            {
                if (templateTypeInfo.IsVerificationSupported)
                {
                    return true;
                }
            }
            return false;
        }

        private static bool HasExtraction(NBBiometricsContext context)
        {
            foreach (NBBiometricsTemplateTypeInfo templateTypeInfo in context.SupportedTemplateTypes)
            {
                if (templateTypeInfo.IsExtractionSupported)
                {
                    return true;
                }
            }
            return false;
        }

        private NBBiometricsContext CreateContext(NBDevice device)
        {
            try
            {
                return new NBBiometricsContext(device);
            }
            catch (Exception ex)
            {
                MessageBox.Show(string.Format("Failed to connect to create biometrics context. Reason: {0}", ex.Message), Text,
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }
        }

        #endregion
    }
}
