﻿using System;
using System.Collections.Generic;
using System.Data.SQLite;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Windows.Forms;
using NextBiometrics.Biometrics;

namespace NBBiometricsSampleCS.Misc
{
    public class Database
    {
        #region Public types

        public class Template
        {
            public int Id { get; private set; }
            public string Name { get; private set; }
            public NBBiometricsTemplate Data { get; private set; }

            public Template(int id, string name, NBBiometricsTemplate data)
            {
                Id = id;
                Name = name;
                Data = data;
            }

            public override string ToString()
            {
                return string.Format("{0} ({1}, {2})", Name, Data.Position, Data.Type);
            }
        }

        #endregion

        #region Private fields

        private readonly List<Template> _database = new List<Template>();

        #endregion

        #region Public properties

        public IReadOnlyCollection<Template> Templates
        {
            get { return _database; }
        }

        public NBBiometricsContext BiometricsContext { get; set; }

        #endregion

        #region Public constructor

        public Database(NBBiometricsContext context)
        {
            BiometricsContext = context;
        }

        #endregion

        #region Private methods

        private string GetDatabaseFileName()
        {
            var localAppData = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData);
            var folderName = Path.Combine(localAppData, "NEXT Biometrics");
            var subFolderName = Path.Combine(folderName, "SDKs");
            if (!Directory.Exists(subFolderName))
            {
                Directory.CreateDirectory(subFolderName);
            }
            var databaseName = string.Format("NBBiometricsSample_{0}.db", BiometricsContext.AlgorithmInfo.Id);
            return Path.Combine(subFolderName, databaseName);
        }

        private string GetConnectionString(string fileName)
        {
            return string.Format("Data Source={0};Version=3;", fileName);
        }

        private void Create(string fileName)
        {
            SQLiteConnection.CreateFile(fileName);
            var connectionString = GetConnectionString(fileName);
            using (var connection = new SQLiteConnection(connectionString))
            {
                connection.Open();

                using (var command = connection.CreateCommand())
                {
                    command.CommandText =
                        "CREATE TABLE 'Properties' ('Name' VARCHAR(255) PRIMARY KEY," +
                        "'Value' VARCHAR(255) NOT NULL);";
                    command.ExecuteNonQuery();
                    command.CommandText = string.Format("INSERT INTO 'Properties' ('Name', 'Value') VALUES ('SampleName', '{0}');" +
                                                        "INSERT INTO 'Properties' ('Name', 'Value') VALUES ('SampleVersion', '{1}');" +
                                                        "INSERT INTO 'Properties' ('Name', 'Value') VALUES ('AlgorithmId', '{2}');" +
                                                        "INSERT INTO 'Properties' ('Name', 'Value') VALUES ('AlgorithmVersion', '{3}');", Application.ProductName, Application.ProductVersion, BiometricsContext.AlgorithmInfo.Id, BiometricsContext.AlgorithmInfo.Version.ToString());
                    command.ExecuteNonQuery();
                }

                using (var command = connection.CreateCommand())
                {
                    command.CommandText =
                        "CREATE TABLE 'Templates' ('Id' INTEGER PRIMARY KEY AUTOINCREMENT, 'Name' TEXT NOT NULL UNIQUE," +
                        "'Type' INTEGER NOT NULL, 'Position' INTEGER NOT NULL, 'Data' BLOB NOT NULL);";
                    command.ExecuteNonQuery();
                }
            }
        }

        private IEnumerable<KeyValuePair<string, string>> GetProperties(string fileName)
        {
            var connectionString = GetConnectionString(fileName);
            using (var connection = new SQLiteConnection(connectionString))
            {
                connection.Open();
                using (var command = connection.CreateCommand())
                {
                    command.CommandText =
                        "SELECT t.'Name', t.'Value' FROM 'Properties' t";
                    var list = new List<KeyValuePair<string, string>>();
                    using (var reader = command.ExecuteReader())
                    {
                        while (reader.Read())
                        {
                            var name = reader.GetString(0);
                            var value = reader.GetString(1);
                            list.Add(new KeyValuePair<string, string>(name, value));
                        }
                    }
                    return list;
                }
            }
        }

        private Template AddTemplate(string name, NBBiometricsTemplate template)
        {
            var fileName = GetDatabaseFileName();
            var connectionString = GetConnectionString(fileName);
            using (var connection = new SQLiteConnection(connectionString))
            {
                connection.Open();

                using (var command = connection.CreateCommand())
                {
                    command.CommandText = "INSERT INTO 'Templates' ('Name', 'Type', 'Position', 'Data') VALUES (@param1, @param2, @param3, @param4); SELECT last_insert_rowid() FROM 'Templates'";
                    command.Parameters.Add(new SQLiteParameter("@param1", name));
                    command.Parameters.Add(new SQLiteParameter("@param2", (int)template.Type));
                    command.Parameters.Add(new SQLiteParameter("@param3", (int)template.Position));
                    command.Parameters.Add(new SQLiteParameter("@param4", BiometricsContext.SaveTemplate(template)));
                    return new Template(Convert.ToInt32(command.ExecuteScalar()), name, template);
                }
            }
        }

        private void DeleteTemplate(Template template)
        {
            var fileName = GetDatabaseFileName();
            var connectionString = GetConnectionString(fileName);
            using (var connection = new SQLiteConnection(connectionString))
            {
                connection.Open();

                using (var command = connection.CreateCommand())
                {
                    command.CommandText = "DELETE FROM 'Templates' WHERE Id = @param1";
                    command.Parameters.AddWithValue("@param1", template.Id);
                    command.ExecuteNonQuery();
                }
            }
        }

        #endregion

        #region Public methods

        public void Initialize()
        {
            var fileName = GetDatabaseFileName();
            if (!File.Exists(fileName))
            {
                Create(fileName);
            }
            var properties = GetProperties(fileName);
            var algorithmId = properties.Where(x => x.Key == "AlgorithmId").Select(x => x.Value).SingleOrDefault();
            if (string.IsNullOrEmpty(algorithmId) || int.Parse(algorithmId) != BiometricsContext.AlgorithmInfo.Id)
            {
                throw new ApplicationException("Loaded database was generated by another algorithm");
            }
            var connectionString = GetConnectionString(fileName);
            using (var connection = new SQLiteConnection(connectionString))
            {
                connection.Open();
                using (var command = connection.CreateCommand())
                {
                    command.CommandText =
                        "SELECT t.'Id', t.'Name', t.'Type', t.'Position', t.'Data' FROM 'Templates' t";
                    using (var reader = command.ExecuteReader())
                    {
                        while (reader.Read())
                        {
                            var id = reader.GetInt32(0);
                            var name = reader.GetString(1);
                            var type = (NBBiometricsTemplateType)reader.GetInt32(2);
                            var data = (byte[])reader.GetValue(4);
                            _database.Add(new Template(id, name, BiometricsContext.LoadTemplate(type, data)));
                        }
                    }
                }
            }
        }

        public void Terminate()
        {
        }

        public void Add(string name, NBBiometricsTemplate template)
        {
            var record = _database.FirstOrDefault(x => x.Name == name);
            if (record != null)
                throw new ApplicationException("Template with specified name already exists in the database");
            _database.Add(AddTemplate(name, template));
        }

        public void Remove(Template template)
        {
            var record = _database.First(x => x.Name == template.Name);
            if (record != null)
            {
                DeleteTemplate(record);
                _database.Remove(record);
            }
        }

        #endregion
    }
}
