﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;
using NBBiometricsSampleCS.Forms;
using NBBiometricsSampleCS.Misc;
using NextBiometrics.Devices;

namespace NBBiometricsSampleCS.Tabs
{
    public partial class CaptureImageTab : BaseTab
    {
        #region Private types

        private enum CaptureType
        {
            Snapshot,
            Continuous,
            OneFingerprint
        }

        #endregion

        #region Private fields

        private NBDeviceScanFormatInfo _currentScanFormat;

        private CaptureType _captureType;
        private int _timeout;
        private bool _doForce;
        private ImageForm _imageForm;
        private static bool _isSpoofDetected = false;
        private const string _spoofDetected = "SpoofDetected";
        private const int _resetAntiSpoof = -1;
        private int _spoofThreshold;
        private const int supportAntiSpoof = 108;
        private const int enableAntiSpoofThreshold = 109;
        private const int enableAntiSpoof = 1;
        private const int disableAntiSpoof = 0;
        private const int defaultAntiSpoofThreshold = 363;

        #endregion

        #region Public constructor

        public CaptureImageTab()
        {
            InitializeComponent();
        }

        #endregion

        #region Public properties

        public override NBDevice Device
        {
            get { return base.Device; }
            set
            {
                base.Device = value;
                if (base.Device == null)
                    ClearInfo();
                else
                {
                    UpdateInfo();
                }
            }
        }

        public override bool IsBusy
        {
            get { return backgroundWorker.IsBusy; }
        }

        #endregion

        #region Public methods

        public override void Cancel()
        {
            if (Device != null && IsBusy)
            {
                buttonCancel.PerformClick();
                while (IsBusy)
                {
                    Application.DoEvents();
                    Thread.Sleep(100);
                }
            }
        }

        #endregion

        #region Private form events

        private void radioButtonSnapshot_CheckedChanged(object sender, EventArgs e)
        {
            checkBoxUseTimeout.Enabled = numericUpDownMiliseconds.Enabled = radioButtonOneFingerprint.Checked;
        }

        private void pictureBoxImage_DoubleClick(object sender, EventArgs e)
        {
            if (_imageForm == null || _imageForm.IsDisposed)
            {
                _imageForm = new ImageForm();
                _imageForm.Show();
            }
        }

        private void buttonStart_Click(object sender, EventArgs e)
        {
            _currentScanFormat = comboBoxFormat.SelectedItem as NBDeviceScanFormatInfo;
            _captureType = radioButtonContinuous.Checked ? CaptureType.Continuous : radioButtonSnapshot.Checked ? CaptureType.Snapshot :
                CaptureType.OneFingerprint;
            _timeout = checkBoxUseTimeout.Checked ? (int)numericUpDownMiliseconds.Value : NBDevice.ScanTimeoutInfinite;
            _doForce = false;

            NBDeviceType tType = Device.Type;
            //Antispoof support for 65200-U model
            if (tType == NBDeviceType.NB65200U)
            {
                Device.SetParameter(supportAntiSpoof, disableAntiSpoof);
                if (checkBoxUseAntiSpoof.Checked && (_currentScanFormat.Format != NBDeviceScanFormat.FormatCustom))
                {
                    _spoofThreshold = (int)numericUpDownThreshold.Value;
                    Device.SetParameter(supportAntiSpoof, enableAntiSpoof);
                    Device.SetParameter(enableAntiSpoofThreshold, _spoofThreshold);
                }
                else if (checkBoxUseAntiSpoof.Checked && (_currentScanFormat.Format == NBDeviceScanFormat.FormatCustom))
                {
                    MessageBox.Show(this, "Please set the valid width x height(300x400) for custom scan format and Enable Antispoof parameter.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    // The Below two lines of code needs to be Enabled in order to support Antispoof for custom-scan format.
                    //Device.SetParameter(enableAntiSpoofParameter, enableAntiSpoof);
                    //Device.SetParameter(enableAntiSpoofThreshold, _spoofThreshold);
                }
            }

            groupBoxCaptureParameters.Enabled = false;
            buttonCancel.Enabled = true;
            buttonSave.Enabled = false;
            if (_captureType == CaptureType.OneFingerprint)
            {
                buttonForce.Enabled = true;
            }
            backgroundWorker.RunWorkerAsync();
            OnBusyChanged();
        }

        private void buttonCancel_Click(object sender, EventArgs e)
        {
            if (backgroundWorker.IsBusy)
            {
                if (_captureType != CaptureType.OneFingerprint)
                {
                    backgroundWorker.CancelAsync();
                }
                else
                {
                    Device.CancelScan();
                }
            }
        }

        private void buttonForce_Click(object sender, EventArgs e)
        {
            _doForce = true;
        }

        private void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            var worker = (BackgroundWorker)sender;
            if (_captureType == CaptureType.Snapshot)
            {
                _isSpoofDetected = false;
                var result = Device.Scan(_currentScanFormat);
                worker.ReportProgress(0, new PreviewInfo(result, Device.GetFingerDetectValue()));
            }
            else if (_captureType == CaptureType.Continuous)
            {
                while (!backgroundWorker.CancellationPending)
                {
                    _isSpoofDetected = false;
                    var result = Device.Scan(_currentScanFormat);
                    worker.ReportProgress(0, new PreviewInfo(result, Device.GetFingerDetectValue()));
                }
            }
            else
            {
                _isSpoofDetected = false;
                var result = Device.ScanEx(_currentScanFormat, _timeout, (o, args) =>
                {
                    worker.ReportProgress(0, new PreviewInfo(args));
                    if (_doForce)
                    {
                        args.Status = NBDeviceScanStatus.Ok;
                    }
                });
                worker.ReportProgress(0, new PreviewInfo(result, 0));
            }
        }

        private void backgroundWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            var preview = e.UserState as PreviewInfo;
            if (preview == null) return;
            labelStatus.Text = preview.Status.ToString();
            labelFingerprintValue.Text = preview.FingerDetectValue.ToString();
            if (_isSpoofDetected)
            {
                labelStatus.Text = _spoofDetected;
            }
            if (preview.Status.ToString().Equals(_spoofDetected))
            {
                _isSpoofDetected = true;
            }
            SetImage(ImageUtils.CreateImage(preview.Format, preview.Image));
        }

        private void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if (e.Error != null)
            {
                MessageBox.Show(this, string.Format("Error occurred while scanning. Reason: {0}", e.Error.Message), "Error information", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            buttonCancel.Enabled = buttonForce.Enabled = false;
            groupBoxCaptureParameters.Enabled = true;
            buttonSave.Enabled = pictureBoxImage.Image != null;
            if (_isSpoofDetected)
            {
                buttonSave.Enabled = false;
            }
            OnBusyChanged();
        }

        private void buttonSave_Click(object sender, EventArgs e)
        {
            ImageUtils.SaveImage(pictureBoxImage.Image);
        }

        #endregion

        #region Private methods

        private void ClearInfo()
        {
            comboBoxFormat.SelectedIndex = -1;
            radioButtonOneFingerprint.Checked = true;
            checkBoxUseTimeout.Checked = true;
            numericUpDownMiliseconds.Value = 10000;
            checkBoxUseAntiSpoof.Checked = false;
            numericUpDownThreshold.Value = defaultAntiSpoofThreshold;
            SetImage(null);
            labelStatus.Text = labelFingerprintValue.Text = string.Empty;
            Enabled = false;
            buttonSave.Enabled = buttonForce.Enabled = buttonCancel.Enabled = false;
            comboBoxFormat.Items.Clear();
        }

        private void UpdateInfo()
        {
            ClearInfo();

            comboBoxFormat.Items.AddRange(Device.SupportedScanFormats);
            if (comboBoxFormat.Items.Count > 0)
                comboBoxFormat.SelectedIndex = 0;
            Enabled = true;
	    //Snapshot and continuous options are not allowed for 2033U module alone.
            NBDeviceType tType = base.Device.Type;
            if (tType == NBDeviceType.NB2033U)
            {
                radioButtonContinuous.Visible = false;
                radioButtonSnapshot.Visible = false;
                radioButtonOneFingerprint.Enabled = false;
            }
            else
            {
                radioButtonContinuous.Visible = true;
                radioButtonSnapshot.Visible = true;
                radioButtonOneFingerprint.Enabled = true;
            }
        }

        private void SetImage(Image image)
        {
            Image toRemove = pictureBoxImage.Image;
            pictureBoxImage.Image = image;
            if (_imageForm != null && !_imageForm.IsDisposed)
            {
                _imageForm.Update(pictureBoxImage.Image);
            }
            if (toRemove != null)
                toRemove.Dispose();
        }

        #endregion
    }
}
