﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using NBBiometricsSampleCS.Forms;
using NBBiometricsSampleCS.Misc;
using NextBiometrics.Biometrics;
using NextBiometrics.Devices;

namespace NBBiometricsSampleCS.Tabs
{
    public partial class ExtractTab : BaseTab
    {
        #region Private fields

        private NBDeviceScanFormatInfo _currentScanFormatInfo;
        private NBBiometricsTemplateTypeInfo _currentTemplateTypeInfo;
        private NBBiometricsFingerPosition _currentFingerPosition;

        private NBBiometricsTemplate _currentTemplate;

        private int _timeout;
        private bool _doForce;
        private static bool _isSpoofDetected = false;
        private const string _spoofDetected = "SpoofDetected";
        private const int _resetAntiSpoof = -1;
        private int _spoofThreshold;
        private const int supportAntiSpoof = 108;
        private const int enableAntiSpoofThreshold = 109;
        private const int enableAntiSpoof = 1;
        private const int disableAntiSpoof = 0;
        private const int defaultAntiSpoofThreshold = 363;

        #endregion

        #region Public constructor

        public ExtractTab()
        {
            InitializeComponent();
        }

        #endregion

        #region Public properties

        public override NBDevice Device
        {
            get { return base.Device; }
            set
            {
                base.Device = value;
                if (base.Device == null)
                    ClearInfo();
                else
                {
                    UpdateInfo();
                }
            }
        }

        public override bool IsBusy
        {
            get
            {
                return backgroundWorker.IsBusy;
            }
        }

        #endregion

        #region Public methods

        public override void Cancel()
        {
            if (IsBusy)
            {
                buttonCancel.PerformClick();
                while (IsBusy)
                {
                    Application.DoEvents();
                    Thread.Sleep(100);
                }
            }
        }

        #endregion

        #region Private form events

        private void buttonStart_Click(object sender, EventArgs e)
        {
            _currentScanFormatInfo = comboBoxFormat.SelectedItem as NBDeviceScanFormatInfo;
            _currentTemplateTypeInfo = ((TemplateTypeInfo) comboBoxTemplateType.SelectedItem).Type;
            _currentFingerPosition = ((PositionInfo) comboBoxFingerprintPosition.SelectedItem).Type;
            _timeout = checkBoxUseTimeout.Checked ? (int)numericUpDownMiliseconds.Value : NBDevice.ScanTimeoutInfinite;
            _doForce = false;

            NBDeviceType tType = Device.Type;
            //Antispoof support for 65200-U model
            if (tType == NBDeviceType.NB65200U)
            {
                Device.SetParameter(supportAntiSpoof, disableAntiSpoof);
                if (checkBoxUseAntiSpoof.Checked && (_currentScanFormatInfo.Format != NBDeviceScanFormat.FormatCustom))
                {
                    _spoofThreshold = (int)numericUpDownThreshold.Value;
                    Device.SetParameter(supportAntiSpoof, enableAntiSpoof);
                    Device.SetParameter(enableAntiSpoofThreshold, _spoofThreshold);
                }
                else if (_currentScanFormatInfo.Format == NBDeviceScanFormat.FormatCustom)
                {
                    MessageBox.Show(this, "Custom Scan format is not supported for extraction.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    return;
                    // The Below two lines of code needs to be Enabled in order to support Antispoof for custom-scan format.
                    //Device.SetParameter(enableAntiSpoofParameter, enableAntiSpoof);
                    //Device.SetParameter(enableAntiSpoofThreshold, _spoofThreshold);
                }
            }

            groupBoxExtractionParameters.Enabled = false;
            buttonCancel.Enabled = true;
            buttonSave.Enabled = false;
            buttonForce.Enabled = true;
            backgroundWorker.RunWorkerAsync();
            OnBusyChanged();
        }

        private void buttonCancel_Click(object sender, EventArgs e)
        {
            if (backgroundWorker.IsBusy)
            {
                Task.Run(() => BiometricsContext.CancelOperation());
            }
        }

        private void buttonForce_Click(object sender, EventArgs e)
        {
            _doForce = true;
        }

        private void RequestSample()
        {
            if (InvokeRequired)
            {
                Invoke(new MethodInvoker(RequestSample));
                return;
            }
            MessageBox.Show(this, "Please lift your fingerprint and press again (new sample requested)", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        private void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            var worker = (BackgroundWorker)sender;

            _isSpoofDetected = false;

            var extractResult = BiometricsContext.Extract(_currentTemplateTypeInfo.TemplateType, _currentFingerPosition, _currentScanFormatInfo, _timeout, (o, args) =>
            {
                _isSpoofDetected = false;
                if (args.BiometricsStatus == NBBiometricsStatus.NeedMoreSamples)
                {
                    RequestSample();
                }
                else
                {
                    worker.ReportProgress(0, new PreviewInfo(args));
                    if (_doForce)
                    {
                        args.ScanStatus = NBDeviceScanStatus.Ok;
                    }
                }
            });
            e.Result = new BiometricInfo(extractResult);
        }

        private void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if (e.Error != null)
            {
                MessageBox.Show(this, string.Format("Error occurred while extracting. Reason: {0}", e.Error.Message), "Error information", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            buttonCancel.Enabled = buttonForce.Enabled = false;
            groupBoxExtractionParameters.Enabled = true;

            labelFingerprintValue.Text = "";
            var biometricInfo = e.Result as BiometricInfo;
            if (biometricInfo != null)
            {
                _currentTemplate = biometricInfo.Template;
                labelStatus.Text = biometricInfo.Status.ToString();
            }
            else
            {
                _currentTemplate = null;
                labelStatus.Text = "";
            }

            buttonSave.Enabled = _currentTemplate != null;
            if (_isSpoofDetected)
            {
                labelStatus.Text = _spoofDetected;
                buttonSave.Enabled = false;
            }
            OnBusyChanged();
        }

        private void backgroundWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            var preview = e.UserState as PreviewInfo;
            if (preview == null) return;
            labelStatus.Text = preview.Status.ToString();
            labelFingerprintValue.Text = preview.FingerDetectValue.ToString();
            if (preview.Status.ToString().Equals(_spoofDetected))
            {
                _isSpoofDetected = true;
            }
            SetImage(ImageUtils.CreateImage(preview.Format, preview.Image));
        }

        private void buttonSave_Click(object sender, EventArgs e)
        {
            if (_currentTemplate != null)
            {
                using (var exportForm = new ExportForm())
                {
                    exportForm.BiometricsContext = BiometricsContext;
                    exportForm.Database = Database;
                    exportForm.Template = _currentTemplate;
                    exportForm.ShowDialog();
                }
            }
        }

        #endregion

        #region Private methods

        private void ClearInfo()
        {
            comboBoxFormat.SelectedIndex = comboBoxTemplateType.SelectedIndex = comboBoxFingerprintPosition.SelectedIndex = -1;
            checkBoxUseTimeout.Checked = true;
            numericUpDownMiliseconds.Value = 10000;
            checkBoxUseAntiSpoof.Checked = false;
            numericUpDownThreshold.Value = defaultAntiSpoofThreshold;
            SetImage(null);
            labelStatus.Text = labelFingerprintValue.Text = string.Empty;
            Enabled = false;
            buttonSave.Enabled = buttonForce.Enabled = buttonCancel.Enabled = false;
            comboBoxFormat.Items.Clear();
            comboBoxTemplateType.Items.Clear();
            comboBoxFingerprintPosition.Items.Clear();
        }

        private void UpdateInfo()
        {
            ClearInfo();

            comboBoxFingerprintPosition.Items.AddRange(Enum.GetValues(typeof(NBBiometricsFingerPosition))
                .Cast<NBBiometricsFingerPosition>().Select(x => new PositionInfo(x)).ToArray());
            if (comboBoxFingerprintPosition.Items.Count > 0)
                comboBoxFingerprintPosition.SelectedIndex = 0;
            if (Device != null)
            {
                comboBoxFormat.Items.AddRange(Device.SupportedScanFormats);
                if (comboBoxFormat.Items.Count > 0)
                    comboBoxFormat.SelectedIndex = 0;
            }
            if (BiometricsContext != null)
            {
                comboBoxTemplateType.Items.AddRange(BiometricsContext.SupportedTemplateTypes
                    .Where(x => x.IsExtractionSupported).Select(x => new TemplateTypeInfo(x)).ToArray());
                if (comboBoxTemplateType.Items.Count > 0)
                    comboBoxTemplateType.SelectedIndex = 0;
            }
            Enabled = true;
        }

        private void SetImage(Image image)
        {
            Image toRemove = pictureBoxImage.Image;
            pictureBoxImage.Image = image;
            //if (_imageForm != null && !_imageForm.IsDisposed)
            //{
            //    _imageForm.Update(pictureBoxImage.Image);
            //}
            if (toRemove != null)
                toRemove.Dispose();
        }

        #endregion
    }
}
