﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using NBBiometricsSampleCS.Misc;
using NextBiometrics.Biometrics;
using NextBiometrics.Devices;

namespace NBBiometricsSampleCS.Tabs
{
    public partial class VerifyTab : BaseTab
    {
        #region Private fields

        private NBDeviceScanFormatInfo _currentScanFormatInfo;
        private SecurityLevelInfo _currentSecurityLevelInfo;
        private Database.Template _currentTemplate;

        private int _timeout;
        private bool _doForce;
        private static bool _isSpoofDetected = false;
        private const string _spoofDetected = "SpoofDetected";
        private const int _resetAntiSpoof = -1;
        private int _spoofThreshold;
        private const int supportAntiSpoof = 108;
        private const int enableAntiSpoofThreshold = 109;
        private const int enableAntiSpoof = 1;
        private const int disableAntiSpoof = 0;
        private const int defaultAntiSpoofThreshold = 363;

        #endregion

        #region Public constructor

        public VerifyTab()
        {
            InitializeComponent();
        }

        #endregion

        #region Public properties

        public override NBDevice Device
        {
            get { return base.Device; }
            set
            {
                base.Device = value;
                RefreshData();
            }
        }

        public override bool IsBusy
        {
            get
            {
                return backgroundWorker.IsBusy;
            }
        }

        #endregion

        #region Public methods

        public override void RefreshData()
        {
            if (Device == null)
                ClearInfo();
            else
                UpdateInfo();
        }

        public override void Cancel()
        {
            if (IsBusy)
            {
                buttonCancel.PerformClick();
                while (IsBusy)
                {
                    Application.DoEvents();
                    Thread.Sleep(100);
                }
            }
        }

        #endregion

        #region Private form events

        private void buttonStart_Click(object sender, EventArgs e)
        {
            _currentScanFormatInfo = comboBoxFormat.SelectedItem as NBDeviceScanFormatInfo;
            _currentSecurityLevelInfo = comboBoxSecurityLevel.SelectedItem as SecurityLevelInfo;;
            _currentTemplate = comboBoxTemplate.SelectedItem as Database.Template;
            if (_currentTemplate == null)
            {
                MessageBox.Show(this, "Template to verify against is not selected or database is empty, please enroll at least one template and try again.", "Error information", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            _timeout = checkBoxUseTimeout.Checked ? (int)numericUpDownMiliseconds.Value : NBDevice.ScanTimeoutInfinite;
            _doForce = false;

            NBDeviceType tType = Device.Type;
            //Antispoof support for 65200-U model
            if (tType == NBDeviceType.NB65200U)
            {
                Device.SetParameter(supportAntiSpoof, disableAntiSpoof);
                if (checkBoxUseAntiSpoof.Checked && (_currentScanFormatInfo.Format != NBDeviceScanFormat.FormatCustom))
                {
                    _spoofThreshold = (int)numericUpDownThreshold.Value;
                    Device.SetParameter(supportAntiSpoof, enableAntiSpoof);
                    Device.SetParameter(enableAntiSpoofThreshold, _spoofThreshold);
                }
                else if (_currentScanFormatInfo.Format == NBDeviceScanFormat.FormatCustom)
                {
                    MessageBox.Show(this, "Custom Scan format is not supported for Verification.", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    return;
                    // The Below two lines of code needs to be Enabled in order to support Antispoof for custom-scan format.
                    //Device.SetParameter(enableAntiSpoofParameter, enableAntiSpoof);
                    //Device.SetParameter(enableAntiSpoofThreshold, _spoofThreshold);
                }
            }

            groupBoxVerificationParameters.Enabled = false;
            buttonCancel.Enabled = true;
            buttonForce.Enabled = true;
            backgroundWorker.RunWorkerAsync();
            OnBusyChanged();
        }

        private void buttonCancel_Click(object sender, EventArgs e)
        {
            if (backgroundWorker.IsBusy)
            {
                Task.Run(() => BiometricsContext.CancelOperation());
            }
        }

        private void buttonForce_Click(object sender, EventArgs e)
        {
            _doForce = true;
        }

        private void RequestSample()
        {
            if (InvokeRequired)
            {
                Invoke(new MethodInvoker(RequestSample));
                return;
            }
            MessageBox.Show(this, "Please lift your fingerprint and press again (new sample requested)", "Information", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        private void backgroundWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            var worker = (BackgroundWorker)sender;

            _isSpoofDetected = false;

            var verifyResult = BiometricsContext.Verify(_currentTemplate.Data.Type, _currentTemplate.Data.Position, _currentScanFormatInfo, _timeout, (o, args) =>
            {
                _isSpoofDetected = false;
                if (args.BiometricsStatus == NBBiometricsStatus.NeedMoreSamples)
                {
                    RequestSample();
                }
                else
                {
                    worker.ReportProgress(0, new PreviewInfo(args));
                    if (_doForce)
                    {
                        args.ScanStatus = NBDeviceScanStatus.Ok;
                    }
                }
            }, _currentTemplate.Data, _currentSecurityLevelInfo.Type);
            e.Result = new BiometricInfo(verifyResult);
        }

        private void backgroundWorker_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if (e.Error != null)
            {
                MessageBox.Show(this, string.Format("Error occurred during verification. Reason: {0}", e.Error.Message), "Error information", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            buttonCancel.Enabled = buttonForce.Enabled = false;
            groupBoxVerificationParameters.Enabled = true;

            var biometricInfo = e.Result as BiometricInfo;
            if (biometricInfo != null)
            {
                labelStatus.Text = biometricInfo.Status == NBBiometricsStatus.Ok
                    ? string.Format("Verified with score {0}", biometricInfo.Score)
                    : biometricInfo.Status == NBBiometricsStatus.MatchNotFound ? 
                        string.Format("NOT verified, score: {0}", biometricInfo.Score) : 
                        biometricInfo.Status.ToString();
            }
            else
            {
                labelStatus.Text = "";
            }
            labelFingerprintValue.Text = "";
            if (_isSpoofDetected)
            {
                labelStatus.Text = _spoofDetected;
            }
            OnBusyChanged();
        }

        private void backgroundWorker_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            var preview = e.UserState as PreviewInfo;
            if (preview == null) return;
            labelStatus.Text = preview.Status.ToString();
            labelFingerprintValue.Text = preview.FingerDetectValue.ToString();
            if (preview.Status.ToString().Equals(_spoofDetected))
            {
                _isSpoofDetected = true;
            }
            SetImage(ImageUtils.CreateImage(preview.Format, preview.Image));
        }

        #endregion

        #region Private methods

        private void ClearInfo()
        {
            comboBoxFormat.SelectedIndex = comboBoxTemplate.SelectedIndex = comboBoxSecurityLevel.SelectedIndex = -1;
            checkBoxUseTimeout.Checked = true;
            numericUpDownMiliseconds.Value = 10000;
            checkBoxUseAntiSpoof.Checked = false;
            numericUpDownThreshold.Value = defaultAntiSpoofThreshold;
            SetImage(null);
            labelStatus.Text = labelFingerprintValue.Text = string.Empty;
            Enabled = false;
            buttonForce.Enabled = buttonCancel.Enabled = false;
            comboBoxFormat.Items.Clear();
            comboBoxTemplate.Items.Clear();
            comboBoxSecurityLevel.Items.Clear();
        }

        private void UpdateInfo()
        {
            ClearInfo();

            if (Device != null)
            {
                comboBoxFormat.Items.AddRange(Device.SupportedScanFormats);
                if (comboBoxFormat.Items.Count > 0)
                    comboBoxFormat.SelectedIndex = 0;
            }
            IEnumerable<Database.Template> templates = Database != null ? Database.Templates : new List<Database.Template>();
            if (BiometricsContext != null)
            {
                var templateTypes = BiometricsContext.SupportedTemplateTypes.Where(x =>
                    x.IsVerificationSupported).ToArray();
                templates = templates.Where(x => templateTypes.FirstOrDefault(y => y.TemplateType == x.Data.Type) != null);
            }
            comboBoxTemplate.Items.AddRange(templates.ToArray());
            if (comboBoxTemplate.Items.Count > 0)
                comboBoxTemplate.SelectedIndex = 0;
            var securityLevels = Enum.GetValues(typeof(NBBiometricsSecurityLevel))
                .Cast<NBBiometricsSecurityLevel>().Select(x => new SecurityLevelInfo(x)).ToArray();
            comboBoxSecurityLevel.Items.AddRange(securityLevels);
            if (comboBoxSecurityLevel.Items.Count > 0)
            {
                var normalSecurityLevel =
                    securityLevels.FirstOrDefault(x => x.Type == NBBiometricsSecurityLevel.Normal);
                if (normalSecurityLevel != null)
                    comboBoxSecurityLevel.SelectedItem = normalSecurityLevel;
                else
                    comboBoxSecurityLevel.SelectedIndex = 0;
            }

            Enabled = true;
        }

        private void SetImage(Image image)
        {
            Image toRemove = pictureBoxImage.Image;
            pictureBoxImage.Image = image;
            //if (_imageForm != null && !_imageForm.IsDisposed)
            //{
            //    _imageForm.Update(pictureBoxImage.Image);
            //}
            if (toRemove != null)
                toRemove.Dispose();
        }

        #endregion
    }
}
